'''
* Copyright 2015-2021 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
will be approved by the European Commission - subsequent
versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
writing, software distributed under the Licence is
distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
express or implied.
* See the Licence for the specific language governing
permissions and limitations under the Licence.
'''

from collections import deque

from scipy.ndimage.measurements import center_of_mass as CoM

from .core import *
from .vtkinteractorstyles import CalcamInteractorStyle2D, CalcamInteractorStyle3D
from ..calibration import Calibration, Fitter, ImageUpsideDown
from ..pointpairs import PointPairs
from ..render import render_cam_view
from .. import misc
from ..image_enhancement import enhance_image, scale_to_8bit
from ..movement import manual_movement

max_undo_depth = 20

# Main calcam window class for actually creating calibrations.
class FittingCalib(CalcamGUIWindow):

    def __init__(self, app, parent = None, load_file=None):

        # GUI initialisation
        CalcamGUIWindow.init(self,'fitting_calib.ui',app,parent)

        # Start up with no CAD model
        self.cadmodel = None
        self.calibration = Calibration(cal_type='fit')

        # Set up VTK
        self.qvtkwidget_3d = qt.QVTKRenderWindowInteractor(self.vtkframe_3d)
        self.vtkframe_3d.layout().addWidget(self.qvtkwidget_3d)
        self.interactor3d = CalcamInteractorStyle3D(refresh_callback=self.refresh_3d,viewport_callback=self.update_viewport_info,cursor_move_callback=self.update_cursor_position,newpick_callback=self.new_point_3d,focus_changed_callback=lambda x: self.change_point_focus('3d',x),resize_callback=self.update_vtk_size,pre_move_callback=self.record_undo_state)
        self.qvtkwidget_3d.SetInteractorStyle(self.interactor3d)
        self.renderer_3d = vtk.vtkRenderer()
        self.renderer_3d.SetBackground(0, 0, 0)
        self.qvtkwidget_3d.GetRenderWindow().AddRenderer(self.renderer_3d)
        self.camera_3d = self.renderer_3d.GetActiveCamera()

        self.qvtkwidget_2d = qt.QVTKRenderWindowInteractor(self.vtkframe_2d)
        self.vtkframe_2d.layout().addWidget(self.qvtkwidget_2d)
        self.interactor2d = CalcamInteractorStyle2D(refresh_callback=self.refresh_2d,newpick_callback = self.new_point_2d,cursor_move_callback=self.update_cursor_position,focus_changed_callback=lambda x: self.change_point_focus('2d',x),pre_move_callback=self.record_undo_state)
        self.qvtkwidget_2d.SetInteractorStyle(self.interactor2d)
        self.renderer_2d = vtk.vtkRenderer()
        self.renderer_2d.SetBackground(0, 0, 0)
        self.qvtkwidget_2d.GetRenderWindow().AddRenderer(self.renderer_2d)
        self.camera_2d = self.renderer_2d.GetActiveCamera()

        self.populate_models()



        # Disable image transform buttons if we have no image
        self.image_settings.hide()

        self.tabWidget.setTabEnabled(2,False)
        self.tabWidget.setTabEnabled(3,False)
        self.tabWidget.setTabEnabled(4,False)


        # Callbacks for GUI elements
        self.image_sources_list.currentIndexChanged.connect(self.build_imload_gui)
        self.viewlist.itemSelectionChanged.connect(self.change_cad_view)
        self.camX.valueChanged.connect(self.change_cad_view)
        self.camY.valueChanged.connect(self.change_cad_view)
        self.camZ.valueChanged.connect(self.change_cad_view)
        self.tarX.valueChanged.connect(self.change_cad_view)
        self.tarY.valueChanged.connect(self.change_cad_view)
        self.tarZ.valueChanged.connect(self.change_cad_view)
        self.cam_roll.valueChanged.connect(self.change_cad_view)
        self.camFOV.valueChanged.connect(self.change_cad_view)
        self.load_model_button.clicked.connect(self.load_model)
        self.model_name.currentIndexChanged.connect(self.populate_model_variants)
        self.feature_tree.itemChanged.connect(self.update_checked_features)
        self.load_image_button.clicked.connect(self.load_image)
        self.im_flipud.clicked.connect(self.transform_image)
        self.im_fliplr.clicked.connect(self.transform_image)
        self.im_rotate_button.clicked.connect(self.transform_image)
        self.im_reset.clicked.connect(self.transform_image)
        self.im_y_stretch_button.clicked.connect(self.transform_image)
        self.load_pointpairs_button.clicked.connect(self.load_pointpairs)
        self.fitted_points_checkbox.toggled.connect(self.toggle_reprojected)
        self.overlay_checkbox.toggled.connect(self.update_overlay)
        self.enhance_checkbox.stateChanged.connect(self.toggle_enhancement)
        self.im_define_splitFOV.clicked.connect(self.edit_split_field)
        self.pixel_size_checkbox.toggled.connect(self.update_pixel_size)
        self.pixel_size_box.valueChanged.connect(self.update_pixel_size)
        self.load_chessboard_button.clicked.connect(self.modify_chessboard_constraints)
        self.chessboard_checkbox.toggled.connect(self.toggle_chessboard_constraints)
        self.load_intrinsics_calib_button.clicked.connect(self.modify_intrinsics_calib)
        self.intrinsics_calib_checkbox.toggled.connect(self.toggle_intrinsics_calib)
        self.viewport_load_calib.clicked.connect(self.load_viewport_calib)
        self.action_save.triggered.connect(self.save_calib)
        self.action_save_as.triggered.connect(lambda: self.save_calib(saveas=True))
        self.action_open.triggered.connect(self.load_calib)
        self.action_new.triggered.connect(self.reset)
        self.action_cal_info.triggered.connect(self.show_calib_info)
        self.overlay_colour_button.clicked.connect(self.change_overlay_colour)
        self.comparison_overlay_colour_button.clicked.connect(self.change_comparison_colour)
        self.open_comparison_calib.clicked.connect(self.select_comparison_calib)
        self.comparison_overlay_checkbox.toggled.connect(self.update_overlay)
        self.overlay_type.currentIndexChanged.connect(self.update_overlay)
        self.comparison_overlay_type.currentIndexChanged.connect(self.update_overlay)
        self.comparison_overlay_opacity_slider.valueChanged.connect(self.change_comparison_colour)

        self.control_sensitivity_slider.valueChanged.connect(lambda x: self.interactor3d.set_control_sensitivity(x*0.01))
        self.rmb_rotate.toggled.connect(self.interactor3d.set_rmb_rotate)
        self.interactor3d.set_control_sensitivity(self.control_sensitivity_slider.value()*0.01)

        self.del_pp_button.clicked.connect(self.remove_current_pointpair)
        self.clear_points_button.clicked.connect(self.clear_pointpairs)

        self.points_undo_button.clicked.connect(self.pointpairs_undo)

        self.overlay_opacity_slider.valueChanged.connect(self.change_overlay_colour)

        # Set up some keyboard shortcuts
        # It is done this way in 3 lines per shortcut to avoid segfaults on some configurations
        sc = qt.QShortcut(qt.QKeySequence(qt.QKeySequence.Delete),self)
        sc.setContext(qt.Qt.WindowShortcut)
        sc.activated.connect(self.remove_current_pointpair)

        sc = qt.QShortcut(qt.QKeySequence("Ctrl+F"),self)
        sc.setContext(qt.Qt.WindowShortcut)
        sc.activated.connect(self.do_fit)

        sc = qt.QShortcut(qt.QKeySequence("Ctrl+P"),self)
        sc.setContext(qt.Qt.WindowShortcut)
        sc.activated.connect(self.toggle_reprojected)

        sc = qt.QShortcut(qt.QKeySequence(qt.QKeySequence.Undo),self)
        sc.setContext(qt.Qt.WindowShortcut)
        sc.activated.connect(self.pointpairs_undo)

        # Odds & sods
        self.pixel_size_box.setSuffix(u' \u00B5m')

        self.viewport_calibs = DodgyDict()
        # Populate image sources list and tweak GUI layout for image loading.
        self.imload_inputs = {}
        self.image_load_options.layout().setColumnMinimumWidth(0,100)

        self.image_sources = self.config.get_image_sources()
        index = -1
        for i,imsource in enumerate(self.image_sources):
            self.image_sources_list.addItem(imsource.display_name)
            if imsource.display_name == self.config.default_image_source:
                index = i

        self.image_sources_list.setCurrentIndex(index)

        self.chessboard_pointpairs = []

        self.point_pairings = []
        self.selected_pointpair = None

        self.fitters = []
        self.fit_settings_widgets = []

        self.fit_overlay = None
        self.comp_overlay = None

        self.chessboard_history = []

        self.fit_results = []

        self.filename = None

        self.waiting_pointpairs = None

        self.pointpairs_history = deque(maxlen=max_undo_depth)

        self.n_points = [ [0,0] ] # One list per sub-field, which is [extrinsics_data, intrinsics_data]

        self.intrinsics_calib = None

        self.fit_initted = False

        self.overlay_opacity_slider.setValue(self.config.main_overlay_colour[3]*100)
        self.overlay_appearance_controls.hide()
        self.comparison_overlay_appearance.hide()

        # Start the GUI!
        self.show()
        self.interactor2d.init()
        self.interactor3d.init()
        self.qvtkwidget_3d.GetRenderWindow().GetInteractor().Initialize()
        self.qvtkwidget_2d.GetRenderWindow().GetInteractor().Initialize()
        self.interactor3d.on_resize()

        if load_file is not None:
            self.load_calib(load_file)


    def modify_intrinsics_calib(self):
        loaded_calib = self.object_from_file('calibration')

        if loaded_calib is not None:
            self.intrinsics_calib = loaded_calib
            if self.intrinsics_calib_checkbox.isChecked():
                self.intrinsics_calib_checkbox.setChecked(False)

            self.intrinsics_calib_checkbox.setChecked(True)
            self.unsaved_changes = True


    def toggle_intrinsics_calib(self,on):

        if on and self.intrinsics_calib is None:
            self.modify_intrinsics_calib()
            if self.intrinsics_calib is None:
                self.intrinsics_calib_checkbox.setChecked(False)
        self.reset_fit()
        self.update_pointpairs()
        self.update_n_points()
        self.unsaved_changes = True


    def select_comparison_calib(self):

        cal = self.object_from_file('calibration')

        if cal is not None:

            if all([vm is None for vm in cal.view_models]):
                raise UserWarning('The selected calibration file does not contain a camera model so cannot be used for comparison.')

            curr_shape = self.calibration.geometry.get_display_shape()
            comp_shape = cal.geometry.get_display_shape()
            if curr_shape != comp_shape:
                raise UserWarning('The selected calibration has different image dimensions ({:d} x {:d}) to the current calibration ({:d} x {:d}), so cannot be used to compare!'.format(comp_shape[0],comp_shape[1],curr_shape[0],curr_shape[1]))

            self.comp_calib = cal
            self.comparison_overlay_checkbox.setEnabled(True)
            self.comparison_overlay_checkbox.setChecked(True)
            self.comparison_name.setText(('...' + cal.filename[-25:]) if len(cal.filename) > 28 else cal.filename)

        elif self.comp_calib is None:
            self.comparison_overlay_checkbox.setChecked(False)
            self.comparison_overlay_checkbox.setEnabled(False)



            oversampling = int(np.ceil(min(1000 / np.array(self.calibration.geometry.get_display_shape()))))

            fname = cal.filename

            self.statusbar.showMessage('Rendering wireframe overlay...')
            self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
            self.app.processEvents()

            orig_colours = self.cadmodel.get_colour()
            if self.comparison_overlay_type.currentIndex() == 0:
                self.cadmodel.set_wireframe(True)
            self.cadmodel.set_colour((1,1,1))
            self.comp_overlay = render_cam_view(self.cadmodel,cal,transparency=True,verbose=False,aa=2,oversampling=oversampling)
            self.cadmodel.set_colour(orig_colours)
            self.cadmodel.set_wireframe(False)

            self.statusbar.clearMessage()
            self.app.restoreOverrideCursor()
            self.comparison_overlay_checkbox.setEnabled(True)
            self.comparison_overlay_checkbox.setChecked(True)


    def on_close(self):
        self.qvtkwidget_2d.close()
        self.qvtkwidget_3d.close()


    def reset(self,keep_cadmodel=False):

        self.fit_initted = False

        if not keep_cadmodel:
            self.tabWidget.setTabEnabled(2,False)
            if self.cadmodel is not None:
                self.cadmodel.remove_from_renderer(self.renderer_3d)
                self.cadmodel.unload()
                self.feature_tree.blockSignals(True)
                self.feature_tree.clear()
                self.feature_tree.blockSignals(False)
                self.cadmodel = None

        self.clear_pointpairs()

        self.interactor2d.set_image(None)

        self.calibration = Calibration(cal_type='fit')

        self.image_settings.hide()

        self.comp_overlay = None
        self.comparison_overlay_checkbox.setChecked(False)
        self.comparison_overlay_checkbox.setEnabled(False)
        self.comparison_name.setText('No comparison calibration loaded')

        self.tabWidget.setTabEnabled(3,False)
        self.tabWidget.setTabEnabled(4,False)

        self.filename = None
        self.setWindowTitle('Calcam Calibration Tool (Point Fitting)')

        self.chessboard_pointpairs = []
        self.chessboard_checkbox.setChecked(False)
        self.intrinsics_calib = None
        self.intrinsics_calib_checkbox.setChecked(False)

        self.refresh_2d()
        self.refresh_3d()
        self.unsaved_changes = False


    def change_overlay_colour(self):

        if self.sender() is self.overlay_opacity_slider:
            new_colour = self.config.main_overlay_colour[:3]

        else:
            old_colour = self.config.main_overlay_colour
            new_colour = self.pick_colour(init_colour=old_colour)

        if new_colour is not None:

            new_colour = new_colour + [self.overlay_opacity_slider.value() / 100]
            self.config.main_overlay_colour = new_colour

            self.update_overlay()


    def change_comparison_colour(self):

        if self.sender() is self.comparison_overlay_opacity_slider:
            new_colour = self.config.second_overlay_colour[:3]

        else:
            old_colour = self.config.second_overlay_colour
            new_colour = self.pick_colour(init_colour=old_colour)

        if new_colour is not None:

            new_colour = new_colour + [self.comparison_overlay_opacity_slider.value()/100]
            self.config.second_overlay_colour = new_colour

            self.update_overlay()


    def reset_fit(self,subview=None):

        if subview is None:
            if not any(self.calibration.view_models):
                return
            self.calibration.view_models = [None] * self.calibration.n_subviews
            self.calibration.history['fit'] = [None] * self.calibration.n_subviews
        else:
            if self.calibration.view_models[subview] is None:
                return
            self.calibration.view_models[subview] = None
            self.calibration.history['fit'][subview] = None

        self.fit_overlay = None
        self.update_fit_results(show_points = self.fitted_points_checkbox.isChecked())

        self.unsaved_changes = True


    def update_cursor_position(self,cursor_id,new_position):

        self.unsaved_changes = True
        self.update_cursor_info()
        self.update_pointpairs()


    def record_undo_state(self):

        pp = self.calibration.pointpairs
        hist = self.calibration.history['pointpairs']
        try:
            coords3d = self.interactor3d.get_cursor_coords(self.point_pairings[self.selected_pointpair][0])
        except Exception:
            coords3d = None

        self.pointpairs_history.append((pp,hist,coords3d))
        self.points_undo_button.setEnabled(True)


    def new_point_2d(self,im_coords):

        self.record_undo_state()

        if self.selected_pointpair is not None:
            if self.point_pairings[self.selected_pointpair][1] is None:
                self.point_pairings[self.selected_pointpair][1] = self.interactor2d.add_active_cursor(im_coords)
                self.interactor2d.set_cursor_focus(self.point_pairings[self.selected_pointpair][1])
                self.update_pointpairs()
                self.update_n_points()
                return

        self.point_pairings.append( [None,self.interactor2d.add_active_cursor(im_coords)] )
        self.interactor2d.set_cursor_focus(self.point_pairings[-1][1])
        self.interactor3d.set_cursor_focus(None)
        self.selected_pointpair = len(self.point_pairings) - 1
        self.update_cursor_info()

        self.update_n_points()


    def new_point_3d(self,coords):

        self.record_undo_state()
        if self.selected_pointpair is not None:
            if self.point_pairings[self.selected_pointpair][0] is None:
                self.point_pairings[self.selected_pointpair][0] = self.interactor3d.add_cursor(coords)
                self.interactor3d.set_cursor_focus(self.point_pairings[self.selected_pointpair][0])
                self.update_pointpairs()
                self.update_n_points()
                return

        self.point_pairings.append( [self.interactor3d.add_cursor(coords),None] )
        self.interactor3d.set_cursor_focus(self.point_pairings[-1][0])
        self.interactor2d.set_cursor_focus(None)
        self.selected_pointpair = len(self.point_pairings) - 1
        self.update_cursor_info()

        self.update_n_points()




    def change_point_focus(self,sender,new_focus):

        if self.selected_pointpair is not None:
            if None in self.point_pairings[self.selected_pointpair]:
                if self.point_pairings[self.selected_pointpair][0] is not None:
                    self.interactor3d.remove_cursor(self.point_pairings[self.selected_pointpair][0])
                if self.point_pairings[self.selected_pointpair][1] is not None:
                    self.interactor2d.remove_active_cursor(self.point_pairings[self.selected_pointpair][1])
                self.point_pairings.remove(self.point_pairings[self.selected_pointpair])

        if sender == '3d':
            for i,pointpair in enumerate(self.point_pairings):
                if pointpair[0] == new_focus:
                    self.interactor2d.set_cursor_focus(pointpair[1])
                    self.selected_pointpair = i
        elif sender == '2d':
            for i,pointpair in enumerate(self.point_pairings):
                if pointpair[1] == new_focus:
                    self.interactor3d.set_cursor_focus(pointpair[0])
                    self.selected_pointpair = i

        self.update_cursor_info()


    def update_cursor_info(self):

        if self.selected_pointpair is not None:
            if self.point_pairings[self.selected_pointpair][0] is not None:
                object_coords = self.interactor3d.get_cursor_coords(self.point_pairings[self.selected_pointpair][0])
            else:
                object_coords = None

            if self.point_pairings[self.selected_pointpair][1] is not None:
                image_coords = self.interactor2d.get_cursor_coords(self.point_pairings[self.selected_pointpair][1])
            else:
                image_coords = None

            info_string = ''

            if object_coords is not None and self.cadmodel is not None:
                info_string = info_string + '<span style=" text-decoration: underline;">CAD Point<br></span>' + self.cadmodel.format_coord(object_coords).replace('\n','<br>') + '<br><br>'

            if image_coords is not None:
                info_string = info_string + '<span style=" text-decoration: underline;">Image Point(s)</span><br>'

                for i,point in enumerate(image_coords):
                    if point is not None:
                        if sum(x is not None for x in image_coords) > 1:
                            info_string = info_string + self.calibration.subview_names[i] + ': '.replace(' ','&nbsp;')

                        info_string = info_string + '( {:.0f} , {:.0f} ) px'.format(point[0],point[1]).replace(' ','&nbsp;')

                        if i < len(image_coords) - 1:
                            info_string = info_string + '<br>'


            self.del_pp_button.setEnabled(True)

        else:
            info_string = 'No selection'
            self.del_pp_button.setEnabled(False)

        self.point_info_text.setText(info_string)



    def on_model_load(self):
        # Enable the other tabs!
        self.tabWidget.setTabEnabled(2,True)
        self.update_fit_results()
        #self.tabWidget.setTabEnabled(2,True)
        #self.tabWidget.setTabEnabled(3,True)


    def on_load_image(self,newim):

        imshape = newim['image_data'].shape[1::-1]

        # If we already have an image in the calibration, check if the new one is the same size.
        try:
            # If the new image is the same size, assume all its metadata are the same as the existing image.
            if imshape == tuple(self.calibration.geometry.get_display_shape()) or imshape == tuple(self.calibration.geometry.get_original_shape()):

                if self.calibration.geometry.get_display_shape() == self.calibration.geometry.get_original_shape():
                    if 'coords' not in newim:
                        newim['coords'] = 'display'
                elif imshape == tuple(self.calibration.geometry.get_display_shape()):
                    newim['coords'] = 'display'
                else:
                    newim['coords'] = 'original'

                newim['subview_names'] = self.calibration.subview_names
                newim['subview_mask'] = self.calibration.get_subview_mask(coords=newim['coords'])
                newim['transform_actions'] = self.calibration.geometry.transform_actions
                newim['pixel_aspect'] = self.calibration.geometry.pixel_aspectratio
                newim['pixel_size'] = self.calibration.pixel_size

            else:
                # If the new image is the wrong size, we have to start a new calibration.
                msg = 'The selected image has different dimensions ({:d}x{:d}) to the current one so cannot be loaded without starting a new calibration.<br><br>Do you want to start a new calibration based on the loaded image?'.format(imshape[0],imshape[1])
                reply = qt.QMessageBox.question(self, 'Start new calibration?', msg, qt.QMessageBox.Yes, qt.QMessageBox.Cancel)

                if reply == qt.QMessageBox.Cancel:
                    return
                elif reply == qt.QMessageBox.Yes:
                    self.reset()

            old_image = self.calibration.get_image(coords='Display')

        # A TypeError signifies no image in the existing calibration
        except TypeError:
            old_image = None
            pass

        if newim['pixel_size'] is not None:
            self.pixel_size_checkbox.setChecked(True)
            self.pixel_size_box.setValue(newim['pixel_size']*1e6)
        else:
            self.pixel_size_checkbox.setChecked(False)

        self.calibration.set_image( scale_to_8bit(newim['image_data']) , newim['source'],subview_mask = newim['subview_mask'], transform_actions = newim['transform_actions'],coords=newim['coords'],subview_names=newim['subview_names'],pixel_aspect=newim['pixel_aspect'],pixel_size=newim['pixel_size'],offset=newim['image_offset'] )

        self.interactor2d.set_image(self.calibration.get_image(coords='Display'),n_subviews = self.calibration.n_subviews,subview_lookup = self.calibration.subview_lookup)

        oshape = self.calibration.geometry.get_original_shape()
        self.detector_window = (int(self.calibration.geometry.offset[0]),int(self.calibration.geometry.offset[1]),int(oshape[0]),int(oshape[1]))

        self.image_settings.show()
        if self.enhance_checkbox.isChecked():
            self.enhance_checkbox.setChecked(False)
            self.enhance_checkbox.setChecked(True)

        if self.overlay_checkbox.isChecked():
            self.overlay_checkbox.setChecked(False)
            self.overlay_checkbox.setChecked(True)
        elif self.comparison_overlay_checkbox.isChecked():
            self.comparison_overlay_checkbox.setChecked(False)
            self.comparison_overlay_checkbox.setChecked(True)

        if not self.fit_initted:
            self.init_fitting()

        self.unsaved_changes = True

        self.update_image_info_string(newim['image_data'],self.calibration.geometry)

        if old_image is not None and len(self.point_pairings) > 0:
            self.app.restoreOverrideCursor()
            dialog = qt.QMessageBox(self)
            dialog.setStandardButtons(qt.QMessageBox.Yes | qt.QMessageBox.No)
            dialog.setWindowTitle('Adjust points all at once?')
            dialog.setTextFormat(qt.Qt.RichText)
            dialog.setText('Would you like to open the image movement tool to adjust the calibration points all at once?')
            dialog.setIcon(qt.QMessageBox.Question)
            retcode = dialog.exec()

            if retcode == dialog.Yes:
                pos_lim = np.array(self.calibration.geometry.get_display_shape()) - 0.5
                pp_to_remove = []
                movement = manual_movement(old_image,self.calibration.get_image(coords='Display'),parent_window=self)
                if movement is not None:
                    self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
                    for pp in self.point_pairings:
                        orig_coords = self.interactor2d.get_cursor_coords(pp[1])
                        new_coords = [None] * len(orig_coords)
                        for subview in range(len(orig_coords)):
                            if orig_coords[subview] is not None:
                                new_coords[subview] = movement.ref_to_moved_coords(*orig_coords[subview])
                                if np.all( np.array(new_coords[subview]) >= 0) and np.all( np.array(new_coords[subview]) < pos_lim):
                                    self.interactor2d.set_cursor_coords(pp[1],new_coords[subview],subview=subview)
                                else:
                                    new_coords[subview] = None
                        if all(p is None for p in new_coords):
                            pp_to_remove.append(pp)

                    for pp in pp_to_remove:
                        if pp[0] is not None:
                            self.interactor3d.remove_cursor(pp[0])
                        if pp[1] is not None:
                            self.interactor2d.remove_active_cursor(pp[1])

                        r_index = self.point_pairings.index(pp)
                        if self.selected_pointpair == r_index:
                            if len(self.point_pairings) > 1:
                                self.selected_pointpair = (self.selected_pointpair - 1) % len(self.point_pairings)
                                self.interactor2d.set_cursor_focus(self.point_pairings[self.selected_pointpair][1])
                                self.interactor3d.set_cursor_focus(self.point_pairings[self.selected_pointpair][0])
                            else:
                                self.selected_pointpair = None
                                self.interactor3d.set_cursor_focus(None)
                                self.interactor2d.set_cursor_focus(None)

                        self.point_pairings.remove(pp)

                    self.app.restoreOverrideCursor()
                    self.init_fitting()


    def change_fit_params(self,fun,state):

        fun(state)
        self.fit_enable_check()


    def init_fitting(self,reset_fit=True,reset_options=False):

        self.fit_initted = False

        if reset_fit:
            self.reset_fit()

        # Build the GUI to show fit options, according to the number of fields.
        current_tab = self.subview_tabs.currentIndex()
        self.subview_tabs.clear()

        # List of settings widgets (for showing / hiding when changing model)
        self.perspective_settings = []
        self.fit_settings_widgets = []
        self.fisheye_settings = []
        self.fit_buttons = []
        self.fit_results = []

        if reset_options or len(self.fitters) != self.calibration.n_subviews:
            self.fitters = []

        self.fit_results_widgets = []
        self.view_to_fit_buttons = []

        for field in range(self.calibration.n_subviews):

            if reset_options or len(self.fitters) != self.calibration.n_subviews:
                self.fitters.append(Fitter())

            new_tab = qt.QWidget()
            new_layout = qt.QVBoxLayout()

            options_groupbox = qt.QGroupBox('Fit Options')
            options_layout = qt.QGridLayout()

            # Selection of model
            widgetlist = [qt.QRadioButton('Rectilinear Lens'),qt.QRadioButton('Fisheye Lens')]

            if int(cv2.__version__[0]) < 3:
                widgetlist[1].setEnabled(False)
                widgetlist[1].setToolTip('Requires OpenCV 3')

            widgetlist[0].toggled.connect(self.change_dist_model)
            widgetlist[1].toggled.connect(self.change_dist_model)
            sub_widget = qt.QWidget()
            sub_layout = qt.QHBoxLayout()
            sub_widget.setLayout(sub_layout)
            sub_layout.addWidget(widgetlist[0])
            sub_layout.addWidget(widgetlist[1])
            sub_layout.setContentsMargins(0,0,0,0)
            options_layout.addWidget(sub_widget)


            # Settings for perspective model
            #---------------------------------
            self.perspective_settings.append( qt.QWidget() )
            perspective_settings_layout = qt.QVBoxLayout()
            perspective_settings_layout.setContentsMargins(0,0,0,0)
            self.perspective_settings[-1].setLayout(perspective_settings_layout)

            widgetlist = widgetlist + [qt.QCheckBox('Disable k1'),qt.QCheckBox('Disable k2'),qt.QCheckBox('Disable k3')]

            sub_widget = qt.QWidget()
            sub_layout = qt.QHBoxLayout()
            sub_widget.setLayout(sub_layout)
            sub_layout.addWidget(widgetlist[-3])
            sub_layout.addWidget(widgetlist[-2])
            sub_layout.addWidget(widgetlist[-1])
            sub_layout.setContentsMargins(0,0,0,0)
            perspective_settings_layout.addWidget(sub_widget)

            widgetlist[-3].setChecked(self.fitters[field].fixk1)
            widgetlist[-3].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_k1,state))
            widgetlist[-2].setChecked(self.fitters[field].fixk2)
            widgetlist[-2].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_k2,state))
            widgetlist[-1].setChecked(self.fitters[field].fixk3)
            widgetlist[-1].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_k3,state))
            widgetlist.append(qt.QCheckBox('Disable tangential distortion'))
            perspective_settings_layout.addWidget(widgetlist[-1])
            widgetlist[-1].setChecked(self.fitters[field].disabletangentialdist)
            widgetlist[-1].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_tangential,state))
            widgetlist.append(qt.QCheckBox('Fix Fx = Fy'))
            widgetlist[-1].setChecked(self.fitters[field].fixaspectratio)
            widgetlist[-1].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_aspect,state))
            perspective_settings_layout.addWidget(widgetlist[-1])


            # ------- End of perspective settings -----------------

            # Settings for fisheye model
            #---------------------------------
            self.fisheye_settings.append( qt.QWidget() )
            fisheye_settings_layout = qt.QVBoxLayout()
            fisheye_settings_layout.setContentsMargins(0,0,0,0)
            self.fisheye_settings[-1].setLayout(fisheye_settings_layout)

            widgetlist = widgetlist + [qt.QCheckBox('Disable k1'),qt.QCheckBox('Disable k2'),qt.QCheckBox('Disable k3'),qt.QCheckBox('Disable k4')]

            sub_widget = qt.QWidget()
            sub_layout = qt.QGridLayout()
            sub_widget.setLayout(sub_layout)
            sub_layout.addWidget(widgetlist[-4],0,0)
            sub_layout.addWidget(widgetlist[-3],0,1)
            sub_layout.addWidget(widgetlist[-2],0,2)
            sub_layout.addWidget(widgetlist[-1],1,0)
            sub_layout.setContentsMargins(0,0,0,0)
            fisheye_settings_layout.addWidget(sub_widget)

            widgetlist[-4].setChecked(self.fitters[field].fixk1)
            widgetlist[-4].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_k1,state))
            widgetlist[-3].setChecked(self.fitters[field].fixk2)
            widgetlist[-3].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_k2,state))
            widgetlist[-2].setChecked(self.fitters[field].fixk3)
            widgetlist[-2].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_k3,state))
            widgetlist[-1].setChecked(self.fitters[field].fixk4)
            widgetlist[-1].toggled.connect(lambda state,field=field: self.change_fit_params(self.fitters[field].fix_k4,state))

            for widgetno in [-4,-3,-2,-1]:
                widgetlist[widgetno].toggled.connect(self.fit_enable_check)

            # ------- End of fisheye settings -----------------


            options_layout.addWidget(self.perspective_settings[-1])
            options_layout.addWidget(self.fisheye_settings[-1])

            options_groupbox.setLayout(options_layout)

            fit_button = qt.QPushButton('Do Fit')

            fit_button.clicked.connect(self.do_fit)
            #fit_button.setEnabled(False)
            options_layout.addWidget(fit_button)
            self.fit_buttons.append(fit_button)

            self.fit_settings_widgets.append(widgetlist)

            new_layout.addWidget(options_groupbox)

            results_groupbox = qt.QGroupBox('Fit Results')
            results_layout = qt.QGridLayout()

            self.fit_results.append(results_groupbox)
            results_groupbox.setHidden(True)

            model = self.fitters[field].model
            widgetlist[0].setChecked(model == 'rectilinear')
            widgetlist[1].setChecked(model == 'fisheye')


            # Build GUI to show the fit results, according to the number of fields.

            widgets = [ qt.QLabel('Fit RMS residual = ') , qt.QLabel('Parameter names'),  qt.QLabel('Parameter values'), qt.QPushButton('Set CAD view to match fit')]
            self.view_to_fit_buttons.append(widgets[-1])
            widgets[1].setAlignment(qt.Qt.AlignRight)
            widgets[3].clicked.connect(self.set_fit_viewport)
            results_layout.addWidget(widgets[0],0,0,1,-1)
            results_layout.addWidget(widgets[1],1,0)
            results_layout.addWidget(widgets[2],1,1)
            results_layout.addWidget(widgets[3],2,0,1,-1)
            self.fit_results_widgets.append(widgets)
            results_layout.setColumnMinimumWidth(0,90)
            results_groupbox.setLayout(results_layout)

            new_layout.addWidget(results_groupbox)

            vspace = qt.QSpacerItem(20, 0, qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
            new_layout.addItem(vspace)

            new_tab.setLayout(new_layout)
            self.subview_tabs.addTab(new_tab,self.calibration.subview_names[field])


            for fitter in self.fitters:
                fitter.set_image_shape(self.calibration.geometry.get_display_shape())

            self.tabWidget.setTabEnabled(3,True)
            self.tabWidget.setTabEnabled(4,True)

        if current_tab < self.subview_tabs.count():
            self.subview_tabs.setCurrentIndex(current_tab)

        self.fit_initted = True



    def transform_image(self,data):

        # First, back up the point pair locations in original coordinates.
        orig_pointpairs = self.calibration.geometry.display_to_original_pointpairs(self.calibration.pointpairs)

        self.pointpairs_history.clear()
        self.points_undo_button.setEnabled(False)

        for i in range(len(self.chessboard_pointpairs)):
            self.chessboard_pointpairs[i][0] = self.calibration.geometry.display_to_original_image(self.chessboard_pointpairs[i][0])
            self.chessboard_pointpairs[i][1] = self.calibration.geometry.display_to_original_pointpairs(self.chessboard_pointpairs[i][1])

        if self.sender() is self.im_flipud:
            self.calibration.geometry.add_transform_action('flip_up_down')

        elif self.sender() is self.im_fliplr:
            self.calibration.geometry.add_transform_action('flip_left_right')

        elif self.sender() is self.im_rotate_button:
            self.calibration.geometry.add_transform_action('rotate_clockwise_{:d}'.format(self.im_rotate_angle.value()))

        elif self.sender() is self.im_y_stretch_button:
            self.calibration.geometry.set_pixel_aspect(self.im_y_stretch_factor.value(),absolute=False)

        elif self.sender() is self.im_reset:
            self.calibration.geometry.set_transform_actions([])
            self.calibration.geometry.set_pixel_aspect(1)

        if self.overlay_checkbox.isChecked():
            self.overlay_checkbox.setChecked(False)

        if self.comparison_overlay_checkbox.isChecked():
            self.comparison_overlay_checkbox.setChecked(False)

        if self.fitted_points_checkbox.isChecked():
            self.fitted_points_checkbox.setChecked(False)


        # Update the image and point pairs
        self.interactor2d.set_image(self.calibration.get_image(coords='Display'),n_subviews = self.calibration.n_subviews,subview_lookup=self.calibration.subview_lookup)
        if orig_pointpairs is not None:
            self.load_pointpairs(pointpairs = self.calibration.geometry.original_to_display_pointpairs(orig_pointpairs),history=self.calibration.history['pointpairs'],force_clear=True,clear_fit=False)

        for i in range(len(self.chessboard_pointpairs)):
            self.chessboard_pointpairs[i][0] = self.calibration.geometry.original_to_display_image(self.chessboard_pointpairs[i][0])
            self.chessboard_pointpairs[i][1] = self.calibration.geometry.original_to_display_pointpairs(self.chessboard_pointpairs[i][1])

        if self.enhance_checkbox.isChecked():
            self.enhance_checkbox.setChecked(False)
            self.enhance_checkbox.setChecked(True)


        self.update_image_info_string(self.calibration.get_image(),self.calibration.geometry)
        self.init_fitting()
        self.unsaved_changes = True


    def load_pointpairs(self,data=None,pointpairs=None,src=None,history=None,force_clear=None,clear_fit=True,include_in_undo=True):

        if include_in_undo:
            self.record_undo_state()

        if pointpairs is None:
            pointpairs = self.object_from_file('pointpairs')

        if pointpairs is not None:

            if pointpairs.n_subviews != self.calibration.n_subviews:
                raise UserWarning('These point pairs cannot be used with the current image because they contain a different number of sub-views ({:d} in point pairs vs. {:d} in image).'.format(pointpairs.n_subviews,self.calibration.n_subviews))

            try:
                history = pointpairs.history
            except AttributeError:
                pass
            try:
                src = pointpairs.src
            except AttributeError:
                pass

            if history is None and src is None:
                raise Exception('History or source of the loaded point pairs must be specified!')

            self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
            self.fitted_points_checkbox.setChecked(False)
            self.overlay_checkbox.setChecked(False)

            if (self.pointpairs_clear_before_load.isChecked() or force_clear) and force_clear != False:
                self.clear_pointpairs(include_in_undo=False)

            for i in range(len(pointpairs.object_points)):

                cursorid_2d = None
                for j in range(len(pointpairs.image_points[i])):
                    if pointpairs.image_points[i][j] is not None:
                        if np.min(pointpairs.image_points[i][j]) >= 0 and pointpairs.image_points[i][j][0] < self.calibration.geometry.get_display_shape()[0] and pointpairs.image_points[i][j][1] < self.calibration.geometry.get_display_shape()[1]:
                            if cursorid_2d is None:
                                cursorid_2d = self.interactor2d.add_active_cursor(pointpairs.image_points[i][j])
                            else:
                                self.interactor2d.add_active_cursor(pointpairs.image_points[i][j],add_to=cursorid_2d)

                if cursorid_2d is not None:
                    cursorid_3d = self.interactor3d.add_cursor(pointpairs.object_points[i])
                    self.point_pairings.append([cursorid_3d,cursorid_2d])

            self.update_pointpairs(src=src,history=history,clear_fit=clear_fit)
            self.update_n_points()
            self.update_cursor_info()
            self.app.restoreOverrideCursor()
            self.unsaved_changes = True



    def toggle_reprojected(self,show=None):

        if show is None:
            if self.fitted_points_checkbox.isEnabled():
                self.fitted_points_checkbox.setChecked(not self.fitted_points_checkbox.isChecked())

        elif show:
            self.overlay_checkbox.setChecked(False)
            points = self.calibration.pointpairs.object_points
            projected = self.calibration.project_points(points)
            for point_list in projected:
                for point in point_list:
                    self.interactor2d.add_passive_cursor(point)
        else:
            self.interactor2d.clear_passive_cursors()


    def remove_current_pointpair(self):

        if self.selected_pointpair is not None:

            self.record_undo_state()

            pp_to_remove = self.point_pairings.pop(self.selected_pointpair)

            if len(self.point_pairings) > 0:
                self.selected_pointpair = (self.selected_pointpair - 1) % len(self.point_pairings)
                self.interactor2d.set_cursor_focus(self.point_pairings[self.selected_pointpair][1])
                self.interactor3d.set_cursor_focus(self.point_pairings[self.selected_pointpair][0])
            else:
                self.selected_pointpair = None
                self.interactor3d.set_cursor_focus(None)
                self.interactor2d.set_cursor_focus(None)

            if pp_to_remove[0] is not None:
                self.interactor3d.remove_cursor(pp_to_remove[0])
            if pp_to_remove[1] is not None:
                self.interactor2d.remove_active_cursor(pp_to_remove[1])

            self.update_cursor_info()
            self.update_pointpairs()
            self.update_n_points()

    def pointpairs_undo(self):

        try:
            prev_pointpairs,pp_history,selected_coords = self.pointpairs_history.pop()
            if len(self.pointpairs_history) == 0:
                self.points_undo_button.setEnabled(False)
        except IndexError:
            raise UserWarning('Reached the end of undo history - nothing more to undo.')

        self.load_pointpairs(pointpairs=prev_pointpairs,history=pp_history,force_clear=True,include_in_undo=False)


        for cid in self.interactor3d.cursors.keys():
            if self.interactor3d.get_cursor_coords(cid) == selected_coords:
                self.interactor3d.set_cursor_focus(cid)
                self.change_point_focus('3d',cid)
                break

    def clear_pointpairs(self,include_in_undo=True):

        if include_in_undo:
            self.record_undo_state()

        self.interactor3d.set_cursor_focus(None)
        self.interactor2d.set_cursor_focus(None)
        self.selected_pointpair = None

        for pp in self.point_pairings:
            if pp[0] is not None:
                self.interactor3d.remove_cursor(pp[0])
            if pp[1] is not None:
                self.interactor2d.remove_active_cursor(pp[1])

        self.point_pairings = []

        self.reset_fit()
        self.update_n_points()


    def fit_enable_check(self):

        # This avoids raising errors if this function is called when we have no
        # fit options GUI.
        if len(self.fit_settings_widgets) < self.calibration.n_subviews:
            return


        # Check whether or not we have enough points to enable the fit button.
        for i in range(self.calibration.n_subviews):
            enable = True

            # We need at least 4 extrinsics points and at least as many total points as free parameters
            if self.n_points[i][0] < 4 or np.sum(self.n_points[i]) < 6:
                enable = False

            self.fit_buttons[i].setEnabled(enable)
            if enable:
                self.fit_buttons[i].setToolTip('Do fit')
            else:
                self.fit_buttons[i].setToolTip('Cannot fit: more free parameters than point pair data.')



    def update_pointpairs(self,src=None,history=None,clear_fit=True):

        pp = PointPairs()

        for pointpair in self.point_pairings:
            if pointpair[0] is not None and pointpair[1] is not None:
                pp.add_pointpair(self.interactor3d.get_cursor_coords(pointpair[0]) , self.interactor2d.get_cursor_coords(pointpair[1]) )

        if clear_fit and self.calibration.pointpairs is not None:
            for subview,same in enumerate(pp == self.calibration.pointpairs):
                if not same:
                    self.reset_fit(subview=subview)

        if pp.get_n_points() > 0:

            self.calibration.set_pointpairs(pp,src=src,history=history)

            for subview in range(self.calibration.n_subviews):
                self.fitters[subview].set_pointpairs(self.calibration.pointpairs,subview=subview)
                self.fitters[subview].clear_intrinsics_pointpairs()

        else:
            self.calibration.set_pointpairs(None)


        # Add the intrinsics constraints
        self.calibration.clear_intrinsics_constraints()

        if self.intrinsics_calib_checkbox.isChecked():
            self.calibration.add_intrinsics_constraints(calibration=self.intrinsics_calib)
            for subview in range(self.calibration.n_subviews):
                self.fitters[subview].add_intrinsics_pointpairs(self.intrinsics_calib.pointpairs,subview=subview)
                for ic in self.intrinsics_calib.intrinsics_constraints:
                    self.fitters[subview].add_intrinsics_pointpairs(ic[1])

        if self.chessboard_checkbox.isChecked():
            for n,chessboard_constraint in enumerate(self.chessboard_pointpairs):
                self.calibration.add_intrinsics_constraints(image=chessboard_constraint[0],im_history=self.chessboard_history[n][0],pointpairs = chessboard_constraint[1],pp_history=self.chessboard_history[n][1])
                for subview in range(self.calibration.n_subviews):
                    self.fitters[subview].add_intrinsics_pointpairs(chessboard_constraint[1],subview=subview)


        self.unsaved_changes = True



    def update_fit_results(self,show_points=True):

        self.overlay_checkbox.setChecked(False)
        self.overlay_checkbox.setEnabled(False)
        self.fitted_points_checkbox.setChecked(False)
        self.fitted_points_checkbox.setEnabled(False)

        if not self.fit_initted:
            return

        for subview in range(self.calibration.n_subviews):

            if self.calibration.view_models[subview] is None:
                self.fit_results[subview].hide()
                continue

            # Put the results in to the GUI
            params = self.calibration.view_models[subview]

            # Get CoM of this field on the chip
            ypx,xpx = CoM( self.calibration.get_subview_mask(coords='Display') == subview)

            # Line of sight at the field centre
            los_centre = params.get_los_direction(xpx,ypx)
            fov = self.calibration.get_fov(subview)

            pupilpos = self.calibration.get_pupilpos(subview=subview)

            widgets = self.fit_results_widgets[subview]

            if self.calibration.view_models[subview].model == 'rectilinear':
                widgets[0].setText( '<b>RMS Fit Residual: {: .1f} pixels<b>'.format(params.reprojection_error) )
                widgets[1].setText( ' : <br>'.join( [  'Pupil position' ,
                                                    'View direction' ,
                                                    'Field of view',
                                                    'Focal length' ,
                                                    'Optical centre' ,
                                                    'Distortion coeff. k1' ,
                                                    'Distortion coeff. k2' ,
                                                    'Distortion coeff. k3' ,
                                                    'Distortion coeff. p1' ,
                                                    'Distortion coeff. p2' ,
                                                    ''
                                                    ] ) )
                if self.calibration.pixel_size is not None:
                    fx = params.cam_matrix[0,0] * self.calibration.pixel_size*1e3
                    fy = params.cam_matrix[1,1] * self.calibration.pixel_size*1e3
                    fl_units = 'mm'
                else:
                    fx = params.cam_matrix[0,0]
                    fy = params.cam_matrix[1,1]
                    fl_units = 'px'


                widgets[2].setText( '<br>'.join( [ '( {: .3f} , {: .3f} , {: .3f} ) m'.format(pupilpos[0],pupilpos[1],pupilpos[2]).replace(' ','&nbsp;') ,
                                                '( {: .3f} , {: .3f} , {: .3f} )'.format(los_centre[0],los_centre[1],los_centre[2]).replace(' ','&nbsp;') ,
                                                '{:.1f}\xb0 x {:.1f}\xb0 '.format(fov[0],fov[1]).replace(' ','&nbsp;') ,
                                                "{0:.1f} {2:s} x {1:.1f} {2:s}".format(fx,fy,fl_units).replace(' ','&nbsp;') ,
                                                "( {: .0f} , {: .0f} )".format(params.cam_matrix[0,2], params.cam_matrix[1,2]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[0][0]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[0][1]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[0][4]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[0][2]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[0][3]).replace(' ','&nbsp;') ,
                                                ''
                                                ] ) )
            elif params.model == 'fisheye':
                widgets[0].setText( '<b>RMS Fit Residual: {: .1f} pixels<b>'.format(params.reprojection_error) )
                widgets[1].setText( ' : <br>'.join( [  'Pupil position' ,
                                                    'View direction' ,
                                                    'Field of view',
                                                    'Focal length' ,
                                                    'Optical centre' ,
                                                    'Distortion coeff. k1' ,
                                                    'Distortion coeff. k2' ,
                                                    'Distortion coeff. k3' ,
                                                    'Distortion coeff. k4' ,
                                                    ''
                                                    ] ) )
                if self.calibration.pixel_size is not None:
                    fx = params.cam_matrix[0,0] * self.calibration.pixel_size*1e3
                    fy = params.cam_matrix[1,1] * self.calibration.pixel_size*1e3
                    fl_units = 'mm'
                else:
                    fx = params.cam_matrix[0,0]
                    fy = params.cam_matrix[1,1]
                    fl_units = 'px'

                widgets[2].setText( '<br>'.join( [ '( {: .3f} , {: .3f} , {: .3f} ) m'.format(pupilpos[0],pupilpos[1],pupilpos[2]).replace(' ','&nbsp;') ,
                                                '( {: .3f} , {: .3f} , {: .3f} )'.format(los_centre[0],los_centre[1],los_centre[2]).replace(' ','&nbsp;') ,
                                                '{:.1f}\xb0 x {:.1f}\xb0 '.format(fov[0],fov[1]).replace(' ','&nbsp;') ,
                                                "{0:.1f} {2:s} x {1:.1f} {2:s}".format(fx,fy,fl_units).replace(' ','&nbsp;') ,
                                                "( {: .0f} , {: .0f} )".format(params.cam_matrix[0,2], params.cam_matrix[1,2]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[0]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[1]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[2]).replace(' ','&nbsp;') ,
                                                "{: 5.4f}".format(params.kc[3]).replace(' ','&nbsp;') ,
                                                ''
                                                ] ) )
            if self.cadmodel is not None:
                widgets[3].setEnabled(True)
                self.overlay_checkbox.setEnabled(True)

            self.fit_results[subview].show()
            self.fitted_points_checkbox.setEnabled(True)

        if self.fitted_points_checkbox.isEnabled() and show_points:
            self.fitted_points_checkbox.setChecked(True)




    def do_fit(self):

        subview = self.subview_tabs.currentIndex()

        # If this was called via a keyboard shortcut, we may be in no position to do a fit.
        if not self.fit_buttons[subview].isEnabled():
            return

        self.update_pointpairs()

        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.fitted_points_checkbox.setChecked(False)
        self.overlay_checkbox.setChecked(False)
        self.fit_overlay = None

        # Do the fit!
        self.statusbar.showMessage('Performing calibration fit...')
        try:
            self.calibration.set_fit(subview, self.fitters[subview].do_fit())
            self.statusbar.clearMessage()
        except ImageUpsideDown:
            self.show_msgbox('The fitted calibration indicates that the images is upside down! Try rotating the image 180 degrees and fitting again.')
            self.app.restoreOverrideCursor()
            return

        self.update_fit_results()

        self.app.restoreOverrideCursor()
        if self.tabWidget.isHidden():
            dialog = qt.QMessageBox(self)
            dialog.setStandardButtons(qt.QMessageBox.Close)
            dialog.setWindowTitle('Calcam - Fit Results')
            dialog.setTextFormat(qt.Qt.RichText)
            dialog.setText(str(self.pointpicker.FitResults).replace('\n','<br>'))
            dialog.setIcon(qt.QMessageBox.Information)
            dialog.exec()

        self.unsaved_changes = True


    def render_overlay_image(self,calibration,wireframe):

        oversampling = int(np.ceil(min(1000/np.array(calibration.geometry.get_display_shape()))))

        self.statusbar.showMessage('Rendering CAD image overlay...')
        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.app.processEvents()

        orig_colours = self.cadmodel.get_colour()

        self.cadmodel.set_wireframe(wireframe)
        self.cadmodel.set_colour((1, 1, 1))
        image = render_cam_view(self.cadmodel, calibration, transparency=True, verbose=False, aa=2,oversampling=oversampling)
        self.cadmodel.set_colour(orig_colours)
        self.cadmodel.set_wireframe(False)

        self.statusbar.clearMessage()
        self.app.restoreOverrideCursor()

        return image



    def update_overlay(self):

        # Clear the existing overlay image to force it to re-render if the user has changed between solid / wireframe
        if self.sender() is self.overlay_type:
            self.fit_overlay = None

        if self.sender() is self.comparison_overlay_type:
            self.comp_overlay = None

        # Show or hide extra controls depending on what overlays are enabled
        self.overlay_appearance_controls.setVisible(self.overlay_checkbox.isChecked())
        self.comparison_overlay_appearance.setVisible(self.comparison_overlay_checkbox.isChecked())

        overlay_ims = []

        if self.overlay_checkbox.isChecked():

            self.fitted_points_checkbox.setChecked(False)

            if self.fit_overlay is None:
                self.fit_overlay = self.render_overlay_image(self.calibration,self.overlay_type.currentIndex() == 0)

                if np.max(self.fit_overlay) == 0:
                    self.show_msgbox('CAD model overlay result is a blank image.','This usually means the fit is wildly wrong.')

            # Apply desired colour (stored as a list in self.config.main_overlay_colour)
            overlay_ims.append( (self.fit_overlay * np.tile(np.array(self.config.main_overlay_colour)[np.newaxis,np.newaxis,:],self.fit_overlay.shape[:2] + (1,))).astype(np.uint8) )


        if self.comparison_overlay_checkbox.isChecked():

            self.comparison_overlay_appearance.show()

            if self.comp_overlay is None:
                self.comp_overlay = self.render_overlay_image(self.comp_calib,self.comparison_overlay_type.currentIndex() == 0)

            # Apply desired colour (stored as a list in self.config.second_overlay_colour)
            overlay_ims.append( (self.comp_overlay * np.tile(np.array(self.config.second_overlay_colour)[np.newaxis, np.newaxis, :],self.fit_overlay.shape[:2] + (1,))).astype(np.uint8) )

        self.interactor2d.set_overlay_image(overlay_ims)

        self.refresh_2d()



    def update_n_points(self):

        self.n_points = []

        chessboard_points = 0
        intcal_points = 0

        msg = ''

        for subview in range(self.calibration.n_subviews):
            if subview > 0:
                msg = msg + '\n'

            npts = 0
            for pp in self.point_pairings:
                if pp[0] is not None and pp[1] is not None:
                    im_coords = self.interactor2d.get_cursor_coords(pp[1])

                    if im_coords[subview] is not None:
                        npts += 1

            self.n_points.append([npts,0])

            if self.calibration.n_subviews > 1:
                msg = msg + '{:s}: {:d} point pairs'.format(self.calibration.subview_names[subview],npts)
            else:
                msg = msg + '{:d} point pairs'.format(npts)


        self.n_pointpairs_text.setText(msg)

        if len(self.point_pairings) > 0:
            self.clear_points_button.setEnabled(1)
        else:
            self.clear_points_button.setEnabled(0)


        for chessboard_constraint in self.chessboard_pointpairs:
            for subview in range(self.calibration.n_subviews):
                if self.chessboard_checkbox.isChecked():
                    self.n_points[subview][1] = self.n_points[subview][1] + chessboard_constraint[1].get_n_points(subview) - 6
                chessboard_points = chessboard_points + chessboard_constraint[1].get_n_points(subview)

        if self.intrinsics_calib is not None:
            for subview in range(self.calibration.n_subviews):
                if self.intrinsics_calib_checkbox.isChecked():
                    self.n_points[subview][1] = self.n_points[subview][1] + self.intrinsics_calib.pointpairs.get_n_points(subview) - 6
                intcal_points = intcal_points + self.intrinsics_calib.pointpairs.get_n_points(subview)

                for intrinsics_constraint in self.intrinsics_calib.intrinsics_constraints:
                    if self.intrinsics_calib_checkbox.isChecked():
                        self.n_points[subview][1] = self.n_points[subview][1] + intrinsics_constraint[1].get_n_points(subview) - 6
                    intcal_points = intcal_points + intrinsics_constraint[1].get_n_points(subview)

        if chessboard_points > 0:
            self.chessboard_checkbox.setText('Chessboard Images ({:d} points)'.format(chessboard_points))
        else:
            self.chessboard_checkbox.setText('Chessboard Images (None Loaded)')

        if intcal_points > 0:
            self.intrinsics_calib_checkbox.setText('Existing Calibration ({:d} points)'.format(intcal_points))
        else:
            self.intrinsics_calib_checkbox.setText('Existing Calibration (None Loaded)')

        self.fit_enable_check()
        self.unsaved_changes = True


    def save_calib(self,saveas=False):

        if self.calibration.subview_mask is None:
            raise UserWarning('Nothing to save! You need to load a camera image to calibrate before you can save anything in this tool.')

        if saveas:
            # Back up then clear the current filename if we want to force a new filename choice.
            orig_filename = self.filename
            self.filename = None

        # Save file dialog box
        if self.filename is None:
            self.filename = self.get_save_filename('calibration')

        if self.filename is not None:

            if self.cadmodel is not None:
                viewport = {'cam_x':self.camX.value(),'cam_y':self.camY.value(),'cam_z':self.camZ.value(),'tar_x':self.tarX.value(),'tar_y':self.tarY.value(),'tar_z':self.tarZ.value(),'fov':self.camFOV.value(),'roll':self.cam_roll.value()}
                self.calibration.cad_config = {'model_name':self.cadmodel.machine_name , 'model_variant':self.cadmodel.model_variant , 'enabled_features':self.cadmodel.get_enabled_features(),'viewport':viewport}

            self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
            self.statusbar.showMessage('Saving...')

            try:
                self.calibration.save(self.filename)
            except PermissionError:
                raise UserWarning('Could not write to {:s}: permission denied.'.format(self.filename))

            self.unsaved_changes = False
            self.action_save.setEnabled(True)
            self.setWindowTitle('Calcam Calibration Tool (Point Fitting) - {:s}'.format(os.path.split(self.filename)[-1][:-4]))
            self.statusbar.clearMessage()
            self.app.restoreOverrideCursor()

        elif saveas:
            # Restore original filename if we didn't end up finishing the Save As...
            self.filename = orig_filename




    def load_calib(self,filename=None):

        try:
            opened_calib = Calibration(filename)
        except:
            opened_calib = self.object_from_file('calibration',maintain_window=False)

        if opened_calib is None:
            return

        if opened_calib._type == 'alignment':
            raise UserWarning('The selected calibration is an alignment calibration and cannot be edited in this tool. Please open it with the alignment calibration editor instead.')
        elif opened_calib._type == 'virtual':
            raise UserWarning('The selected calibration is a virtual calibration and cannot be edited in this tool. Please open it in the virtual calibration editor instead.')

        if opened_calib.cad_config is not None:
            cconfig = opened_calib.cad_config
            if self.cadmodel is not None and self.cadmodel.machine_name == cconfig['model_name'] and self.cadmodel.model_variant == cconfig['model_variant']:
                keep_model = True
            else:
                keep_model = False
        else:
            keep_model = False

        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.reset(keep_cadmodel=keep_model)

        # Basic setup
        self.filename = opened_calib.filename
        self.setWindowTitle('Calcam Calibration Tool (Point Fitting) - {:s}'.format(os.path.split(self.filename)[-1][:-4]))

        # Load the image
        for imsource in self.image_sources:
            if imsource.display_name == 'Calcam Calibration':
                try:
                    self.load_image(newim = imsource.get_image_function(self.filename))
                except Exception as e:
                    if 'does not contain an image' in str(e):
                        self.interactor2d.set_image(opened_calib.get_subview_mask(coords='Display')*0,n_subviews=opened_calib.n_subviews,subview_lookup=opened_calib.subview_lookup)
                    else:
                        raise

        self.calibration = opened_calib

        if not self.fit_initted:
            self.init_fitting(reset_fit=False)

        self.interactor2d.set_subview_lookup(self.calibration.n_subviews,self.calibration.subview_lookup)

        self.chessboard_pointpairs = self.calibration.intrinsics_constraints
        self.chessboard_history = self.calibration.history['intrinsics_constraints']

        if len(self.chessboard_pointpairs) > 0:
            self.chessboard_checkbox.blockSignals(True)
            self.chessboard_checkbox.setChecked(True)
            self.chessboard_checkbox.blockSignals(False)


        # Load the point pairs
        if opened_calib.pointpairs is not None:
            self.load_pointpairs(pointpairs=opened_calib.pointpairs,history=opened_calib.history['pointpairs'],force_clear=False,clear_fit=False)


        # Load the appropriate CAD model, if we know what that is
        if opened_calib.cad_config is not None:
            if keep_model:
                try:
                    self.cadmodel.enable_only(cconfig['enabled_features'])
                except Exception as e:
                    if 'Unknown feature' not in str(e):
                        raise
                self.update_feature_tree_checks()
            else:
                cconfig = opened_calib.cad_config
                load_model = True
                try:
                    name_index = sorted(self.model_list.keys()).index(cconfig['model_name'])
                    self.model_name.setCurrentIndex(name_index)
                    variant_index = self.model_list[ cconfig['model_name'] ][1].index(cconfig['model_variant'])
                    self.model_variant.setCurrentIndex(variant_index)
                except ValueError:
                    self.model_name.setCurrentIndex(-1)
                    load_model=False

                if load_model:
                    try:
                        self.load_model(featurelist=cconfig['enabled_features'])
                    except Exception as e:
                        self.cadmodel = None
                        if 'Unknown feature' not in str(e):
                            raise

            self.camX.setValue(cconfig['viewport']['cam_x'])
            self.camY.setValue(cconfig['viewport']['cam_y'])
            self.camZ.setValue(cconfig['viewport']['cam_z'])
            self.tarX.setValue(cconfig['viewport']['tar_x'])
            self.tarY.setValue(cconfig['viewport']['tar_y'])
            self.tarZ.setValue(cconfig['viewport']['tar_z'])
            self.camFOV.setValue(cconfig['viewport']['fov'])
            self.cam_roll.setValue(cconfig['viewport']['roll'])

        for field in range(len(self.fit_settings_widgets)):
            if opened_calib.view_models[field] is not None:

                if opened_calib.view_models[field].model == 'rectilinear':
                    self.fit_settings_widgets[field][0].setChecked(True)
                    widget_index_start = 2
                    widget_index_end = 6
                elif opened_calib.view_models[field].model == 'fisheye':
                    self.fit_settings_widgets[field][1].setChecked(True)
                    widget_index_start = 7
                    widget_index_end = 10


                for widget in self.fit_settings_widgets[field][widget_index_start:widget_index_end+1]:
                    widget.setChecked(False)
                    if str(widget.text()) in opened_calib.view_models[field].fit_options:
                        widget.setChecked(True)



        if self.calibration.readonly:
            self.action_save.setEnabled(False)
        else:
            self.action_save.setEnabled(True)

        self.update_n_points()
        self.update_fit_results()
        self.pointpairs_history.clear()
        self.points_undo_button.setEnabled(False)
        self.app.restoreOverrideCursor()
        self.unsaved_changes = False


    def toggle_enhancement(self,check_state):

        image = self.calibration.get_image(coords='display')

        # Enable / disable adaptive histogram equalisation
        if self.enhance_checkbox.isChecked():
            image = enhance_image(image)

        self.interactor2d.set_image(image,n_subviews = self.calibration.n_subviews,subview_lookup=self.calibration.subview_lookup,hold_position=True)



    def edit_split_field(self):

        dialog = ImageMaskDialog(self,self.interactor2d.get_image())
        result = dialog.exec()

        if result == 1:
            self.calibration.set_subview_mask(dialog.fieldmask,subview_names=dialog.field_names,coords='Display')
            self.interactor2d.set_subview_lookup(self.calibration.n_subviews,self.calibration.subview_lookup)
            self.init_fitting()
            self.unsaved_changes = True
            self.update_n_points()
            self.pointpairs_history.clear()
            self.points_undo_button.setEnabled(False)

        del dialog



    def change_dist_model(self,choice):

        for field in range(len(self.fit_settings_widgets)):
            if self.sender() == self.fit_settings_widgets[field][0]:
                self.perspective_settings[field].show()
                self.fisheye_settings[field].hide()
                self.fitters[field].set_model('rectilinear')
            elif self.sender() == self.fit_settings_widgets[field][1]:
                self.perspective_settings[field].hide()
                self.fisheye_settings[field].show()
                self.fitters[field].set_model('fisheye')

        self.update_n_points()


    def update_pixel_size(self):

        if self.pixel_size_checkbox.isChecked():
            self.pixel_size_box.setEnabled(True)
            self.calibration.pixel_size = self.pixel_size_box.value() / 1e6
        else:
            self.pixel_size_box.setEnabled(False)
            self.calibration.pixel_size = None

        self.update_fit_results()
        self.unsaved_changes = True


    def set_fit_viewport(self):
        subview = self.view_to_fit_buttons.index(self.sender())

        self.set_view_from_calib(self.calibration,subview)



    def modify_chessboard_constraints(self):

        dialog = ChessboardDialog(self)
        dialog.exec()

        # Resizing the window + 1 pixel then immediately back again after the chessboard dialog is closed
        # is a workaround for Issue #65, the root cause of which is a mystery to me. I shake my fist at VTK.
        size = self.size()
        self.resize(size.width(),size.height()+1)
        self.app.processEvents()
        self.resize(size.width(),size.height())
        self.app.processEvents()

        if dialog.results != []:
            self.chessboard_pointpairs = dialog.results
            point_history = ['Auto-detected corners of {:d}x{:d} square chessboard pattern with {:.1f}mm squares.'.format(dialog.chessboard_squares_x.value(),dialog.chessboard_squares_y.value(),dialog.chessboard_square_size.value()),None]
            self.chessboard_history = [ ('Chessboard image from {:s}, loaded by {:s} on {:s} at {:s}.'.format(fname,misc.username,misc.hostname,misc.get_formatted_time()),point_history) for fname in dialog.filenames]
            if self.chessboard_checkbox.isChecked():
                self.chessboard_checkbox.setChecked(False)
            self.chessboard_checkbox.setChecked(True)




    def toggle_chessboard_constraints(self,on):

        if on and len(self.chessboard_pointpairs) == 0:
            self.modify_chessboard_constraints()
            if len(self.chessboard_pointpairs) == 0:
                self.chessboard_checkbox.setChecked(False)

        self.reset_fit()
        self.update_pointpairs()
        self.update_n_points()
        self.fit_enable_check()



'''
* Copyright 2015-2019 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
will be approved by the European Commission - subsequent
versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
writing, software distributed under the Licence is
distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
express or implied.
* See the Licence for the specific language governing
permissions and limitations under the Licence.
'''
import os

import cv2
import vtk
import numpy as np

from .vtkinteractorstyles import CalcamInteractorStyle2D
from . import qt_wrapper as qt
from .. import movement, misc
from ..image_enhancement import enhance_image
from ..config import CalcamConfig

guipath = os.path.split(os.path.abspath(__file__))[0]


class ImageAlignDialog(qt.QDialog):

    def __init__(self,app,parent,ref_image,new_image,correction=None):

        # GUI initialisation
        if parent is None:
            if app is None:
                raise ValueError('Either parent or app must be provided!')
            else:
                self.app = app
        else:
            self.app = parent.app

        qt.QDialog.__init__(self, parent,qt.Qt.WindowTitleHint | qt.Qt.WindowCloseButtonHint)
        qt.uic.loadUi(os.path.join(guipath,'qt_designer_files','movement_corr_dialog.ui'), self)
        self.setWindowIcon(qt.QIcon(os.path.join(guipath, 'icons', 'calcam.png')))

        # See how big the screen is and open the window at an appropriate size
        if qt.qt_ver < 5:
            available_space = self.app.desktop().availableGeometry(self)
        else:
            available_space = self.app.primaryScreen().availableGeometry()

        # Open the window with same aspect ratio as the screen, and no fewer than 500px tall.
        win_height = min(780,0.75*available_space.height())
        win_width = win_height * available_space.width() / available_space.height()
        self.resize(win_width,win_height)

        self.qvtkwidget_ref = qt.QVTKRenderWindowInteractor(self.ref_frame)
        self.ref_frame.layout().addWidget(self.qvtkwidget_ref)
        self.interactor_ref = CalcamInteractorStyle2D(refresh_callback=self.refresh,newpick_callback = lambda x: self.new_point('ref',x),focus_changed_callback=lambda x: self.change_point_focus('ref',x))
        self.qvtkwidget_ref.SetInteractorStyle(self.interactor_ref)
        self.renderer_ref = vtk.vtkRenderer()
        self.renderer_ref.SetBackground(0, 0, 0)
        self.qvtkwidget_ref.GetRenderWindow().AddRenderer(self.renderer_ref)
        self.camera_ref = self.renderer_ref.GetActiveCamera()

        self.qvtkwidget_new = qt.QVTKRenderWindowInteractor(self.new_frame)
        self.new_frame.layout().addWidget(self.qvtkwidget_new)
        self.interactor_new = CalcamInteractorStyle2D(refresh_callback=self.refresh,newpick_callback = lambda x: self.new_point('new',x),focus_changed_callback=lambda x: self.change_point_focus('new',x))
        self.qvtkwidget_new.SetInteractorStyle(self.interactor_new)
        self.renderer_new = vtk.vtkRenderer()
        self.renderer_new.SetBackground(0, 0, 0)
        self.qvtkwidget_new.GetRenderWindow().AddRenderer(self.renderer_new)
        self.camera_new = self.renderer_new.GetActiveCamera()


        self.interactor_new.cursor_size = 0.06
        self.interactor_ref.cursor_size = 0.06

        # GUI callbacks
        self.image_enhancement_checkbox.toggled.connect(self.toggle_enhancement)
        self.update_homography_button.clicked.connect(self.get_transform)
        self.fitted_points_checkbox.toggled.connect(self.toggle_fitted_points)
        self.overlay_button.pressed.connect(lambda: self.toggle_overlay(True))
        self.overlay_button.released.connect(lambda: self.toggle_overlay(False))
        self.clear_button.clicked.connect(self.remove_all_points)
        self.auto_points_button.clicked.connect(self.find_points)
        self.clear_transform_button.clicked.connect(lambda : self.set_correction(None))
        self.save_button.clicked.connect(self.save)
        self.load_button.clicked.connect(self.open)

        # State initialisation
        self.point_pairings = []
        self.overlay_cursors = [[],[]]
        self.selected_pointpair = None
        self.transform_params.hide()
        self.transform = None


        # Start the GUI!
        self.show()
        self.interactor_ref.init()
        self.interactor_new.init()
        self.qvtkwidget_ref.GetRenderWindow().GetInteractor().Initialize()
        self.qvtkwidget_new.GetRenderWindow().GetInteractor().Initialize()

        self.ref_image = ref_image
        self.ref_image_enhanced = None
        self.new_image = new_image
        self.new_image_enhanced = None

        self.interactor_ref.set_image(ref_image)
        self.interactor_new.set_image(new_image)
        self.refresh()

        self.interactor_ref.link_with(self.interactor_new)
        self.interactor_new.link_with(self.interactor_ref)

        sc = qt.QShortcut(qt.QKeySequence(qt.QKeySequence.Delete),self)
        sc.setContext(qt.Qt.WindowShortcut)
        sc.activated.connect(self.remove_current_point)

        self.config = CalcamConfig()

        self.show()

        if correction is not None:
            self.set_correction(correction,include_points=True)


    def refresh(self):
        self.renderer_ref.Render()
        self.qvtkwidget_ref.update()
        self.renderer_new.Render()
        self.qvtkwidget_new.update()


    def new_point(self,image,position):

        if image == 'ref':
            index = self.interactor_ref.add_active_cursor(position)
            if self.selected_pointpair is not None and self.point_pairings[self.selected_pointpair][0] is None:
                self.point_pairings[self.selected_pointpair][0] = index
            else:
                if self.selected_pointpair is not None and self.point_pairings[self.selected_pointpair][1] is None:
                    self.interactor_ref.remove_active_cursor(self.point_pairings[self.selected_pointpair][0])
                    self.point_pairings.remove(self.point_pairings[self.selected_pointpair])
                self.point_pairings.append([index,None])
                self.selected_pointpair = len(self.point_pairings) - 1

        elif image == 'new':

            index = self.interactor_new.add_active_cursor(position)

            if self.selected_pointpair is not None and self.point_pairings[self.selected_pointpair][1] is None:
                self.point_pairings[self.selected_pointpair][1] = index
            else:
                if self.selected_pointpair is not None and self.point_pairings[self.selected_pointpair][0] is None:
                    self.interactor_new.remove_active_cursor(self.point_pairings[self.selected_pointpair][1])
                    self.point_pairings.remove(self.point_pairings[self.selected_pointpair])
                self.point_pairings.append([None, index])
                self.selected_pointpair = len(self.point_pairings) - 1

        self.interactor_ref.set_cursor_focus(self.point_pairings[self.selected_pointpair][0])
        self.interactor_new.set_cursor_focus(self.point_pairings[self.selected_pointpair][1])

        if self.count_pointpairs() > 2:
            self.update_homography_button.setEnabled(True)
        else:
            self.update_homography_button.setEnabled(False)

        self.refresh()

    def find_points(self):

        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.remove_all_points()

        ref_points,new_points = movement.find_pointpairs(self.ref_image,self.new_image)

        self.app.restoreOverrideCursor()

        if ref_points.shape[0] > 0:

            for npair in range(ref_points.shape[0]):
                self.new_point('ref',ref_points[npair,:])
                self.new_point('new',new_points[npair,:])

        else:
            dialog = qt.QMessageBox(self)
            dialog.setStandardButtons(dialog.Ok)
            dialog.setTextFormat(qt.Qt.RichText)
            dialog.setWindowTitle('Could not auto-detect points')
            dialog.setText('Could not auto-detect corresponding points in these two images.')
            dialog.setInformativeText('You will have to identify points manually')
            dialog.setIcon(qt.QMessageBox.Information)
            dialog.exec()




    def remove_current_point(self):

        if self.selected_pointpair is not None:
            if self.point_pairings[self.selected_pointpair][1] is not None:
                self.interactor_new.remove_active_cursor(self.point_pairings[self.selected_pointpair][1])
            if self.point_pairings[self.selected_pointpair][0] is not None:
                self.interactor_ref.remove_active_cursor(self.point_pairings[self.selected_pointpair][0])
            self.point_pairings.remove(self.point_pairings[self.selected_pointpair])

            if len(self.point_pairings) > 0:
                self.selected_pointpair = len(self.point_pairings) - 1
                self.interactor_ref.set_cursor_focus(self.point_pairings[self.selected_pointpair][0])
                self.interactor_new.set_cursor_focus(self.point_pairings[self.selected_pointpair][1])
            else:
                self.selected_pointpair = None

            if len(self.point_pairings) < 3:
                self.update_homography_button.setEnabled(False)

            self.set_correction(None)


    def remove_all_points(self):
        while self.selected_pointpair is not None:
            self.remove_current_point()

    def change_point_focus(self,sender,id):

        if sender == 'ref':
            test_ind = 0
        elif sender == 'new':
            test_ind = 1

        for pp_ind,pointpair in enumerate(self.point_pairings):
            if pointpair[test_ind] == id:
                self.selected_pointpair = pp_ind
                self.interactor_ref.set_cursor_focus(pointpair[0])
                self.interactor_new.set_cursor_focus(pointpair[1])

        self.refresh()


    def toggle_enhancement(self,onoff):

        if onoff:
            if self.ref_image_enhanced is None:
                self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
                self.ref_image_enhanced = enhance_image(self.ref_image)
                self.new_image_enhanced = enhance_image(self.new_image)
                self.app.restoreOverrideCursor()

            self.interactor_new.set_image(self.new_image_enhanced,hold_position=True)
            self.interactor_ref.set_image(self.ref_image_enhanced,hold_position=True)
        else:
            self.interactor_new.set_image(self.new_image,hold_position=True)
            self.interactor_ref.set_image(self.ref_image,hold_position=True)

        self.refresh()


    def get_transform(self):

        self.fitted_points_checkbox.setChecked(False)
        ref_points = np.zeros((self.count_pointpairs(),2))
        new_points = np.zeros((self.count_pointpairs(),2))

        for i,pp in enumerate(self.point_pairings):
            if i < ref_points.shape[0]:
                ref_points[i,:] = self.interactor_ref.get_cursor_coords(pp[0])[0]
                new_points[i,:] = self.interactor_new.get_cursor_coords(pp[1])[0]

        m = np.matrix(cv2.cv2.estimateAffinePartial2D(new_points, ref_points)[0])

        if m[0,0] is None:
            self.transform = None
            self.fitted_points_checkbox.setEnabled(False)
            self.transform_params.hide()
            dialog = qt.QMessageBox(self)
            dialog.setStandardButtons(dialog.Ok)
            dialog.setTextFormat(qt.Qt.RichText)
            dialog.setWindowTitle('Could not determine transform')
            dialog.setText('Could not determine the image transformation. Try adjusting the points and trying again.')
            dialog.setIcon(qt.QMessageBox.Warning)
            dialog.exec()
            return
        else:
            correction = movement.MovementCorrection(m,self.ref_image.shape[:2],ref_points,new_points,'Created using GUI tool by {:s} on {:s} at {:s}'.format(misc.username,misc.hostname,misc.get_formatted_time()))
            self.set_correction(correction)




    def set_correction(self,correction,include_points=False):

        if correction is None:

            if include_points:
                self.remove_all_points()

            self.transform = None
            self.transform_params.hide()
            self.fitted_points_checkbox.setChecked(False)
            self.fitted_points_checkbox.setEnabled(False)
            self.clear_transform_button.setEnabled(False)
            self.save_button.setEnabled(False)


        else:
            if include_points:
                self.remove_all_points()
                for i in range(correction.ref_points.shape[0]):
                    self.new_point('ref',correction.ref_points[i,:])
                    self.new_point('new',correction.moved_points[i, :])

            self.transform = correction

            ddscore = self.transform.get_ddscore(self.ref_image,self.new_image)

            desc = ' ( {:.1f} , {:.1f} ) pixels<br>'.format(*self.transform.translation)
            desc = desc + ' {:.2f}\u00b0<br>'.format(self.transform.rotation)
            desc = desc + ' {:.3f}<br>'.format(self.transform.scale)
            if ddscore >= 0:
                desc = desc + ' <font color="#006400">{:.2f}</font>'.format(ddscore)
            else:
                desc = desc + ' <font color="#ff0000">{:.2f}</font>'.format(ddscore)

            self.transform_params.setText(desc)
            self.transform_params.show()

            self.fitted_points_checkbox.setEnabled(True)
            self.fitted_points_checkbox.setChecked(True)
            self.clear_transform_button.setEnabled(True)
            self.save_button.setEnabled(True)


    def count_pointpairs(self):

        if self.selected_pointpair is not None and None in self.point_pairings[self.selected_pointpair]:
            return len(self.point_pairings) - 1
        else:
            return len(self.point_pairings)


    def toggle_fitted_points(self,onoff):

        for cursor_id in self.overlay_cursors[0]:
            self.interactor_ref.remove_passive_cursor(cursor_id)
        for cursor_id in self.overlay_cursors[1]:
            self.interactor_new.remove_passive_cursor(cursor_id)
        self.overlay_cursors = [[],[]]

        if onoff:
            for pp in self.point_pairings:
                if pp[1] is not None:
                    oldpt = self.interactor_new.get_cursor_coords(pp[1])[0]
                    fitted_pt = self.transform.moved_to_ref_coords(oldpt[0],oldpt[1])
                    self.overlay_cursors[0].append(self.interactor_ref.add_passive_cursor([fitted_pt[0],fitted_pt[1]]))
                    self.overlay_cursors[1].append(self.interactor_new.add_passive_cursor([fitted_pt[0], fitted_pt[1]]))


    def toggle_overlay(self,onoff):

        if onoff:
            if self.image_enhancement_checkbox.isChecked():
                orig_im = self.new_image_enhanced
            else:
                orig_im = self.new_image

            if self.transform is not None:

                overlay_image =self.transform.warp_moved_to_ref(orig_im)[0]

            else:

                overlay_image = orig_im

            self.interactor_ref.set_overlay_image(overlay_image)
            self.interactor_new.set_overlay_image(overlay_image)

        else:
            self.interactor_ref.set_overlay_image(None)
            self.interactor_new.set_overlay_image(None)

        self.refresh()


    def save(self):

        filename_filter = self.config.filename_filters['movement']
        fext = filename_filter.split('(*')[1].split(')')[0]

        filedialog = qt.QFileDialog(self)
        filedialog.setAcceptMode(filedialog.AcceptSave)

        try:
            filedialog.setDirectory(self.config.file_dirs['movement'])
        except KeyError:
            filedialog.setDirectory(os.path.expanduser('~'))


        filedialog.setFileMode(filedialog.AnyFile)

        filedialog.setWindowTitle('Save As...')
        filedialog.setNameFilter(filename_filter)
        filedialog.exec()

        if filedialog.result() == 1:
            selected_path = str(filedialog.selectedFiles()[0])
            self.config.file_dirs['movement'] = os.path.split(selected_path)[0]
            self.config.save()
            if not selected_path.endswith(fext):
                selected_path = selected_path + fext

            self.transform.save(selected_path)


    def open(self):

        filename_filter = self.config.filename_filters['movement']

        filedialog = qt.QFileDialog(self)
        filedialog.setAcceptMode(filedialog.AcceptOpen)

        try:
            filedialog.setDirectory(self.config.file_dirs['movement'])
        except KeyError:
            filedialog.setDirectory(os.path.expanduser('~'))

        filedialog.setFileMode(filedialog.ExistingFile)

        filedialog.setWindowTitle('Open...')
        filedialog.setNameFilter(filename_filter)

        filedialog.exec()


        if filedialog.result() == 1:
            selected_file = filedialog.selectedFiles()[0]
            self.config.file_dirs['movement'] = os.path.split(selected_file)[0]
            self.config.save()

            try:
                correction = movement.MovementCorrection.load(selected_file)
            except Exception as e:
                return

            self.set_correction(correction,include_points=True)


'''
* Copyright 2015-2019 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''


'''
Geometry matrix module for Calcam.

Written by Scott Silburn & James Harrison.
'''
import multiprocessing
import copy
import time
import json
import os
import random

import numpy as np
import scipy.sparse
import scipy.io

import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.patches import Polygon as PolyPatch
from matplotlib.collections import PatchCollection
import matplotlib.path as mplpath

try:
    import meshpy.triangle
    meshpy_err = None
except Exception as e:
    meshpy_err = '{:}'.format(e)

from . import config
from . import misc
from .coordtransformer import CoordTransformer
from .io import ZipSaveFile


class PoloidalVolumeGrid:
    '''
    Class for representing tomographic reconstruction grids with
    polygonal grid cells in the R, Z plane. Quantities are assumed to be
    uniform within the volume of each grid cell. Grid cells can be arbitrary 
    polygons but all cells in the grid must all have the same number of 
    sides.
    
    Grids can be constructed by directly instantiating this class with the 
    following parameters, or alternatively convenience functions such as 
    :func:`calcam.gm.squaregrid` are provided for easily creating various types of grid.

    Parameters:
        
        vertices (numpy.ndarray)    : (N_verts x 2) array of floats containing the \
                                      (R,Z) coordinates of the grid cell vertices.
                                
        cells (numpy.ndarray)       : (N_cells x N_verts_per_cell) array of integers \
                                      specifying which vertices (indexes in to the vertices \
                                      array) define each grid cell. For each cell (array row), \
                                      the vertices must be listed in order around the \
                                      cell perimeter (in either direction).
                                
        wall_contour (numpy.ndaray) : Nx2 array containing the R,Z wall contour of the machine. \
                                      If provided, this is used for plotting purposes.
                                   
        src (str)                   : Human readable string describing where the grid came from, \
                                      for data provenance purposes.
                                   
    '''
    
    def __init__(self,vertices,cells,wall_contour=None,src=None):

        self.vertices = vertices.copy()
        self.cells = cells.copy()
        self.wall_contour = wall_contour.copy()
        
        if src is None:
            self.history = 'Created by {:s} on {:s} at {:s}'.format(misc.username,misc.hostname,misc.get_formatted_time())
        elif 'by' in src and 'on' in src and 'at' in src:
            self.history = src
        else:
            self.history = src + ' by {:s} on {:s} at {:s}'.format(misc.username,misc.hostname,misc.get_formatted_time())

        self._validate_grid()
        self._build_edge_list()
        self._cull_unused_verts()

        self.gridtype = 'Polyogn Cell Grid'


    @property
    def n_cells(self):
        '''
		The number of cells in the grid.
        '''
        return self.cells.shape[0]

    @property
    def n_segments(self):
        '''
        The number of line segments (cell sides) in the grid.
        '''
        return self.segments.shape[0]


    @property
    def n_vertices(self):
        '''
		The number of vertices in the grid.
        '''
        return self.vertices.shape[0]


    @property
    def extent(self):
        '''
        A 4-element tuple of floats containing the \
        R,Z extent of the grid in metres (Rmin,Rmax,Zmin,Zmax)
        '''
        rmin = self.vertices[:,0].min()
        rmax = self.vertices[:,0].max()
        zmin = self.vertices[:,1].min()
        zmax = self.vertices[:,1].max()

        return rmin,rmax,zmin,zmax




    def get_cell_intersections(self,ray_start,ray_end,plot=False):
        '''
        Get the intersections of a ray, i.e. a straight line
        in 3D space, with the grid cell boundaries.
        
        Parameters:
            
            ray_start (sequence) : 3-element sequence containing the X,Y,Z coordinates \
                                   of the ray's start position.

            ray_end (sequence)   : 3-element sequence containing the X,Y,Z coordinates \
                                   of the ray's end position.
                                   
            plot (bool)          : If set to True, the function will also plot the ray \
                                   in R,Z and red circles at each intersection location. \
                                   Mainly intended for debugging and algorithm demonstration.
                        
        Returns:
            
            tuple   : Intersection information:
                      
                      * NumPy array containing the lengths along the ray where intersections \
                        with grid cell boundaries lines were found (in order from start to end of the ray).
                         
                      * list of lists containing the indices of the grid cell(s) which were involved \
                        in  each intersection. Each intersection will be associated with 1 or more grid cells.

        '''

        # Turn off some NumPy warnings because we will inevitably
        # have some dividing by zero and such in here, but it's harmless.
        with np.errstate(divide='ignore',invalid='ignore'):
            
            ray_start = np.array(ray_start)
            ray_end = np.array(ray_end)
            
            ray_length = np.sqrt( np.sum( (ray_end - ray_start)**2 ) )
            
            # Parametric coefficients for ray
            pax,pay,paz = ray_start
            dpx,dpy,dpz = (ray_end - ray_start) / ray_length
            
            # Parametric coefficients for grid segments
            lar = self.vertices[self.segments[:,0],0]
            laz = self.vertices[self.segments[:,0],1]
            dlr = self.vertices[self.segments[:,1],0] - lar
            dlz = self.vertices[self.segments[:,1],1] - laz
    
            a = -dlz**2*dpx**2 - dlz**2*dpy**2 + dlr**2*dpz**2
            b = 2*dlr*dlz*dpz*lar - 2*dlr**2*dpz*laz - 2*dlz**2*dpx*pax - 2*dlz**2*dpy*pay + 2*dlr**2*dpz*paz
            c = (dlz**2*lar**2 - 2*dlr*dlz*lar*laz + dlr**2*laz**2 - dlz**2*pax**2 - dlz**2*pay**2 + 2*dlr*dlz*lar*paz - 2*dlr**2*laz*paz + dlr**2*paz**2)
            
            # These will be the intersection positions
            t_ray0 = np.zeros(self.n_segments) - 1.
            t_seg0 = np.zeros(self.n_segments) - 1.
            t_ray1 = np.zeros(self.n_segments) - 1.
            t_seg1 = np.zeros(self.n_segments) - 1.
    
            # The magic number!
            d = b**2 - 4*a*c
    
            # d > 0 means two real solutions and hence two intersections
            indx = np.where(d > 0)
            if len(indx) > 0:
                q = -0.5 * (b[indx] + np.sign(b[indx]) * np.sqrt(d[indx]))
                t_ray0[indx] = q/a[indx]
                t_seg0[indx] = (-laz[indx] + paz + dpz * t_ray0[indx])/dlz[indx]
                t_ray1[indx] = c[indx] / q
                t_seg1[indx] = (-laz[indx] + paz + dpz * t_ray1[indx])/dlz[indx]
    
            # d == 0 means one real solution so one intersection
            indx = np.where(d == 0)
            if len(indx) > 0:
                q = -0.5 * (b[indx] + np.sign(b[indx]) * np.sqrt(d[indx]))
                t_ray0[indx] = q / a[indx]
                t_seg0[indx] = (-laz[indx] + paz + dpz * t_ray0[indx])/dlz[indx]
            
            # Special case for exactly horizontal rays.
            indx = np.where(np.abs(dlz) <1e-14)
            if len(indx) > 0:
                t_ray0[indx] = (-paz + laz[indx])/dpz
                hitr = np.sqrt((pax+t_ray0[indx]*dpx)**2+(pay+t_ray0[indx]*dpy)**2)
                t_seg0[indx] = (-lar[indx] + hitr)/dlr[indx]
                
            
            # Valid intersections are ones within the line segment length and 
            # within the ray length
            valid_inds0 = (t_seg0 >= 0.) & (t_seg0 <= 1.) & (t_ray0 >= 0.) & (t_ray0 <= ray_length)
            valid_inds1 = (t_seg1 >= 0.) & (t_seg1 <= 1.) & (t_ray1 >= 0.) & (t_ray1 <= ray_length)
            

            # Full list of intersection distance and segment index
            t_ray = np.concatenate( (t_ray0[valid_inds0],t_ray1[valid_inds1]) )     
            seg_inds = np.arange(self.n_segments,dtype=np.uint32)
            seg_inds = np.concatenate( (seg_inds[valid_inds0],seg_inds[valid_inds1]) )
    
            
            # Sort the intersections by length along the sight-line.
            # Also round t_ray to 9 figures because we'll want to find unique values
            # of it shortly, so round to something well over machine precision.
            sort_order = np.argsort(t_ray)
            t_ray = t_ray[sort_order].round(decimals=12)
            seg_inds = seg_inds[sort_order]
            
            # This will be the output list of cell indices
            cell_inds = []
            
            # Check which cells the intersected line segments belong to.
            for intersection_pos in np.unique( t_ray ):
                
                segs = seg_inds[t_ray == intersection_pos]
                
                cells = set()
                for seg in seg_inds[ t_ray == intersection_pos]:
                    cells.update( np.where(self.cell_sides == seg)[0] )
                
                cell_inds.append(list(cells))
            
            # We only want to return the unique intersection lengths
            t_ray = np.unique(t_ray)

            # If we're asked to, plot the sight line and intersection points.
            if plot:
                
                # Plot the sight line
                ray_dir = (ray_end - ray_start) / ray_length
                l = np.linspace(0,ray_length,int(ray_length/1e-2))
                R = np.sqrt( (ray_start[0] + l*ray_dir[0])**2 + (ray_start[1] + l*ray_dir[1])**2  )
                Z = ray_start[2] + l*ray_dir[2]
                plt.plot(R,Z)
                
                # Plot the intersections
                points3d = np.tile(ray_start[np.newaxis,:],(t_ray.size,1)) + np.tile(t_ray[:,np.newaxis],(1,3)) * np.tile(ray_dir[np.newaxis,:],(t_ray.size,1))
                R = np.sqrt(np.sum(points3d[:,:2]**2,axis=1))
                Z = points3d[:,2]
                plt.plot(R,Z,'ro')


        return t_ray,cell_inds




    def plot(self,data=None,clim=None,cmap=None,line_colour=(0,0,0),cell_linewidth=None,cblabel='',axes=None):
        '''
        Either plot a given data vector on the grid, or if no data vector is given,
        plot the grid itself.

        Parameters:

            data (array)                    : Data to plot on the grid. Must be a 1D array with as many elements \
                                              as there are grid cells. If not given, only the grid structure is plotted.

            clim (sequence or None)         : 2 element sequence giving the colourmap limits for the data [min,max]. \
                                              If set to None, the data min and max are used.

            cmap (str or None)              : Name of the matplotlib colourmap to use to display the data. If not given, \
                                              matplotlib's default colourmap will be used.

            line_colour (sequence or None)  : 3-element sequence of values between 0 and 1 specifying the R,G,B \
                                              colour with which to draw the wall contour and grid cell boundaries. \
                                              If set to None, the wall and cell boundaries are not drawn.

            cell_linewidth (float)          : Line width to use to show the grid cell boundaries. If set to 0, the grid \
                                              cell boundaries will not be drawn.

            cblabel (str)                   : Label for the data colour bar, if plotting data.

            axes (matplotlib.pyplot.Axes)   : Matplotlib axes on which to plot. If not given, a new figure will be created.


        Returns:
            
            tuple : MatPlotLib objects comprising the plot:
                    
                    * matplotlib.axes.Axes                    : The matplotlib axes containing the plot.
            
                    * matplotlib.collections.PatchCollection  : PatchCollection containing the patches used to show the data and \
                                                                grid cells. This object has useful methods for further adjusting \
                                                                the plot like set_cmap(), set_clim() etc.

                    * list of matplotlib.lines.Line2D         : List of matpltolib line objects making up the wall contour.

        '''

        # If we have some data to plot, validate that we have 1 value per grid cell
        if data is not None:

            data = np.array(data)

            if len(np.squeeze(data).shape) > 1:
                raise ValueError('Data must be a 1D array the same length as the number of grid cells ({:d}).'.format(self.n_cells))

            if data.size != self.n_cells:
                raise ValueError('Data vector is the wrong length! Data has {:d} values but the grid gas {:d} cells.'.format(data.size,self.n_cells))

            cmap = cm.get_cmap(cmap)

            # If we have data, grid cell edges default to off
            if cell_linewidth is None:
                cell_linewidth = 0

        else:
            # If we have no data, grid cell edges default to on
            if cell_linewidth is None:
                cell_linewidth = 0.25


        # Create a matplotlib patch collection to show the grid cells,
        # with the data associated with it if we have data.
        patches = []
        verts_per_cell = self.cells.shape[1]

        for cell_ind in range(self.n_cells):
            xy = np.vstack( [ self.vertices[self.cells[cell_ind,i],:] for i in range(verts_per_cell)] )
            patches.append( PolyPatch( xy, closed=True) )
        pcoll = PatchCollection(patches)
        if data is not None:
            pcoll.set_array(np.squeeze(data))
        else:
            pcoll.set_array(np.zeros(self.n_cells)+np.nan)


        # Set up appearance of the grid cells
        if data is not None:
            pcoll.set_cmap(cmap)
            if clim is not None:
                pcoll.set_clim(clim)

        pcoll.set_edgecolor(line_colour)
        pcoll.set_linewidth(cell_linewidth)


        # If not given an existing set of axes to use, make a new figure.
        if axes is None:
            fig = plt.figure()
            axes = fig.add_subplot(111)
        
        # Plot the grid cells!
        axes.add_collection(pcoll)

        # Plot the wall
        if line_colour is not None and self.wall_contour is not None:
            wall_lines = axes.plot(self.wall_contour[:,0],self.wall_contour[:,1],color=line_colour,linewidth=2,marker=None)
            wall_lines = wall_lines + axes.plot([self.wall_contour[-1,0],self.wall_contour[0,0]],[self.wall_contour[-1,1],self.wall_contour[0,1]],color=line_colour,linewidth=2,marker=None)
        else:
            wall_lines = []

        # Add a colour bar if we have data
        if data is not None:
            plt.colorbar(pcoll,label=cblabel if cblabel is not None else '')

        # Prettification - set appropriate zoom for the grid extent.
        rmin,rmax,zmin,zmax = self.extent
        headroom = 0.1
        axes.set_xlim([rmin - headroom,rmax + headroom])
        axes.set_ylim([zmin - headroom, zmax + headroom])
        axes.set_aspect('equal')
        axes.set_xlabel('R [m]')
        axes.set_ylabel('Z [m]')

        return axes,pcoll,wall_lines


    def interpolate(self,data,r_new,z_new,fill_value=np.nan):
        '''
        Given a vector of data on the grid, return the data values at
        positions (r_new,z_new). Since quantities are assumed uniform 
        within each grid cell, this is a "nearest neighbour" type 
        interpolation where the value returned at new each point is 
        the value of the grid cell that point lies within.

        Note: this is not particularly optimised for speed, but at least 
        provides the functionality for now.

        Parameters:

            data (np.ndarray)  : 1D array containing the data defined on the grid; \
                                 must have the same number of elements as there are \
                                 grid cells.

            r_new (np.ndarray) : Arry of new R positions; must be the same shape as z_new

            z_new (np.ndarray) : Array of new Z positions, must be the same shape as r_new

            fill_value (float) : What value to return for (r_new,z_new) points which lie \
                                 outside the grid. Default is np.nan.

        Returns:

            np.ndarray : Array the same shape as r_new and z_new containing the data \
                        values at positions (r_new,z_new).
        '''

        r_new = np.array(r_new)
        z_new = np.array(z_new)

        if r_new.shape != z_new.shape:
            raise ValueError('r_new and z_new must be the same shape!')

        if data.size != self.n_cells:
            raise ValueError('Data vector has wrong size: data has {:d} values but there are {:d} grid cells!'.format(data.size,self.n_cells))

        orig_shape = r_new.shape

        r_new = r_new.reshape(r_new.size)
        z_new = z_new.reshape(z_new.size)

        cell_inds = []

        data_out = []

        for i in range(r_new.size):

            delta = self.vertices.copy()
            delta[:,0] = delta[:,0] - r_new[i]
            delta[:,1] = delta[:,1] - z_new[i]
            theta = np.arctan2(delta[:,1],delta[:,0])
            theta[theta < 0] = theta[theta < 0] + 2.*3.14159

            theta_delta = np.abs( theta[self.cells] - theta[np.roll(self.cells,shift=-1,axis=1)] )
            theta_delta[theta_delta > 3.141592] = 2*3.141592 - theta_delta[theta_delta > 3.141592]

            theta_tot = theta_delta.sum(axis=1)

            cell_inds = np.where(theta_tot > 2*3.1415)[0]

            if cell_inds.size > 0:
                data_out.append(data[cell_inds].mean())
            else:
                data_out.append(fill_value)

        data_out = np.reshape(np.array(data_out),orig_shape)

        return data_out



    def remove_cells(self,cell_inds):
        '''
        Remove grid cells with the given indices from the grid.
        
        Parameters:
            
            cell_inds (sequence) : Sequence of integers specifying which \
                                   cell indices to remove.
        '''
        cell_inds = np.array(cell_inds).astype(np.uint32)

        self.cells = np.delete(self.cells,cell_inds,axis=0)
        self.cell_sides = np.delete(self.cell_sides,cell_inds,axis=0)

        self._cull_unused_verts()


    def _validate_grid(self):
        # Some code here to check the generation worked OK should go here.
        pass



    def _cull_unused_verts(self):
        '''
        Remove any un-used vertices from the mesh definition.
        '''
        used_vert_inds = set(self.cells.flatten())
        all_vert_inds = set(range(self.vertices.shape[0]))
        used_verts = np.array( sorted( list(all_vert_inds & used_vert_inds) ),dtype=np.uint32)

        # Remove the unused vertices and keep track of vertex indexing
        ind_translation = np.zeros(self.vertices.shape[0],dtype=np.uint32)
        ind_translation[used_verts] = np.arange(used_verts.size,dtype=np.uint32)

        self.vertices = self.vertices[used_verts,:]
        self.cells = ind_translation[self.cells]
        self.segments = ind_translation[self.segments]


    def _build_edge_list(self):
        '''
        Build the list of line segments in the grid
        and which line segments border which grid cells.
        '''
        self.segments = np.zeros((np.prod(self.cells.shape),2),dtype=np.uint32)
        self.cell_sides = np.zeros(self.cells.shape,dtype=np.uint32)

        segment_index = 0
        verts_per_cell = self.cells.shape[1]

        # For each grid cell
        for cell_index in range(self.n_cells):

            # For each side of the cell
            for seg_index in range(verts_per_cell):

                seg = sorted( [self.cells[cell_index][seg_index],self.cells[cell_index][(seg_index + 1) % verts_per_cell]] )

                # Store the index if this line segment is already in the segment list,
                # otherwise add it to the segment list and then store its index.
                try:
                    self.cell_sides[cell_index][seg_index] = np.where( (self.segments[:,0] == seg[0]) & (self.segments[:,1] == seg[1]) )[0][0]
                except IndexError:
                    self.segments[segment_index,:] = seg
                    self.cell_sides[cell_index][seg_index] = segment_index
                    segment_index += 1


        self.segments = self.segments[:segment_index,:]




class GeometryMatrix:
    '''
    Class to represent a scalar geometry matrix and associated metadata.
    
    A new geometry matrix can be created by instantiating this class 
    with the below parameters, or a saved geometry matrix can be loaded
    from disk using the :func:`fromfile()` class method.
    
    The matrix itself can be accessed in the `data` attribute, where it is
    stored as a sparse matrix using the scipy.sprase.csr_matrix class.

    Parameters:

        grid (calcam.gm.PoloidalVolumeGrid)  : Reconstruction grid to use

        raydata (calcam.RayData)             : Ray data for the camera to be inverted

        pixel_order (str)                    : What pixel order to use when flattening \
                                               the 2D image array in to the 1D data vector. \
                                               Default 'C' goes left-to-right, then row-by-row (NumPy default), \
                                               alternatively 'F' goes top-to-bottom, then column-by-column (MATLAB default).
        
        calc_status_callback (callable)      : Callable which takes a single argument, which will be called with \
                                               status updates about the calculation. It will be called with either \
                                               a string for textual status updates or a float from 0 to 1 specifying \
                                               the progress of the calculation. By default, status updates are printed \
                                               to stdout.  If set to None, no status updates are issued.

    '''
    def __init__(self,grid,raydata,pixel_order='C',calc_status_callback = misc.LoopProgPrinter().update):

        if grid is not None and raydata is not None:

            if raydata.fullchip:
                if raydata.fullchip == True:
                    raise Exception('Raydata object does not contain information on the image orientation used for ray casting.\n Please set the "fullchip" attribute of the raydata object to either "Display" or "Original".')
                else:
                    self.image_coords = raydata.fullchip
                    self.binning = raydata.binning
                    self.pixel_order = pixel_order
                    if self.image_coords.lower() == 'original':
                        imdims = (np.array(raydata.transform.get_original_shape()[::-1])/self.binning).astype(int)
                    elif self.image_coords.lower() == 'display':
                        imdims = (np.array(raydata.transform.get_display_shape()[::-1])/self.binning).astype(int)
                    self.pixel_mask = np.ones(imdims,dtype=bool)
            else:
                
                self.image_coords = None
                self.binning = None
                self.pixel_order = None
                self.pixel_mask = None

            self.grid = copy.copy(grid)
            '''
            calcam.gm.PoloidalVolumeGrid : The inversion grid associated with the geometry matrix.
            '''    
            
            self.image_geometry = raydata.transform

            self.history = {'los':raydata.history,'grid':grid.history,'matrix':'Created by {:s} on {:s} at {:s}'.format(misc.username,misc.hostname,misc.get_formatted_time())}
            
            # Number of grid cells and sight lines
            n_cells = grid.n_cells
            n_los = raydata.x.size
    
            # Flatten out the ray start and end coords
            ray_start_coords = raydata.ray_start_coords.reshape(-1,3,order=self.pixel_order)
            ray_end_coords = raydata.ray_end_coords.reshape(-1,3,order=self.pixel_order)


            # Multi-threadedly loop over each sight-line in raydata and calculate its matrix row.
            # Store the results as coords + data then build the matrix after, because that is much faster.
            if calc_status_callback is not None:
                calc_status_callback('Calculating geometry matrix elements using {:d} CPUs...'.format(config.n_cpus))
            
            last_status_update = 0.
            
            # We will do the calculation in a random order,
            # purely to get better time remaining estimation.
            inds = list(range(n_los))
            random.shuffle(inds)

            colinds = []
            rowinds = []
            data = []

            with multiprocessing.Pool( config.n_cpus ) as cpupool:
                calc_status_callback(0.)
                for i , row_data in enumerate( cpupool.imap( self._calc_row_volume, np.hstack((ray_start_coords[inds,:],ray_end_coords[inds,:])) , 10 ) ):          
                    rowinds.append(np.zeros(row_data[0].shape,dtype=np.uint32) + inds[i])                    
                    colinds.append(row_data[0])
                    data.append(row_data[1])
                    
                    if time.time() - last_status_update > 1. and calc_status_callback is not None:
                        calc_status_callback(float(i) / n_los)
                        last_status_update = time.time()

            # Build the matrix!
            self.data = scipy.sparse.csr_matrix((np.concatenate(data),(np.concatenate(rowinds),np.concatenate(colinds))),shape=(n_los,n_cells))            
            '''
            scipy.sparse.csr_matrix : The geometry matrix data itself.
            '''

            if calc_status_callback is not None:
                calc_status_callback(1.)
            
            
            # Remove any grid cells + matrix columns which have no sight-line coverage.
            unused_cells = np.where(np.abs(self.data.sum(axis=0)) == 0)[1]
            self.grid.remove_cells(unused_cells)

            used_cols = np.where(self.data.sum(axis=0) > 0)[1]
            self.data = self.data[:,used_cols]

            # Set the pixel mask to exclude pixels which do not contribute to any grid cells and remove corresponding rows.
            if self.pixel_mask is not None:
                used_pixels = np.squeeze(np.array(np.abs(self.data.sum(axis=1))) > 1e-14)
                self.pixel_mask = used_pixels.reshape(imdims,order=self.pixel_order)
                self.data = self.data[used_pixels,:]



    def get_los_coverage(self):
        '''
        Get the number of lines of sight viewing each grid element.
        
        Returns:
            
            numpy.ndarray : Vector with as many elements as there grid cells, \
                            with values being how many sight-lines interact with \
                            that grid element.
        '''
        return np.diff(self.data.tocsc().indptr)



    

    def set_binning(self,binning):
        '''
        Set the level of image binning. Can be used to
        decrease the size of the matrix to reduce memory or
        computation requirements for inversions.
        
        Parameters:
            
            binning (float) : Desired image binning. Must be larger than the \
                              existing binning value.
                              
        '''
        if binning < self.binning:
            raise ValueError('Specified binning is lower than existing binning! The binning can only be increased.')
        elif binning == self.binning:
            return
        else:
            
            if self.image_coords is not None:
                if self.image_coords.lower() == 'display':
                    image_dims = self.image_geometry.get_display_shape()
                elif self.image_coords.lower() == 'original':
                    image_dims = self.image_geometry.get_original_shape()
            else:
                raise Exception('Nope, no worky.')
                
            bin_factor = int(binning / self.binning)

            init_shape = (np.array(image_dims) / self.binning).astype(np.uint32)
            row_inds = np.arange(np.prod(init_shape),dtype=np.uint32)
            row_inds = np.reshape(row_inds,init_shape,order=self.pixel_order)

            px_mask = self.pixel_mask.reshape(self.pixel_mask.size,order=self.pixel_order)
            index_map = np.zeros(np.prod(image_dims)//int(self.binning**2),dtype=np.uint32)
            index_map[px_mask == True] = np.arange(np.count_nonzero(px_mask),dtype=np.uint32)

            self.pixel_mask = _bin_image(self.pixel_mask,bin_factor,bin_func=np.min).astype(bool)

            px_mask = self.pixel_mask.reshape(self.pixel_mask.size,order=self.pixel_order)

            ind_arrays = []
            for colshift in range(bin_factor):
                for rowshift in range(bin_factor):
                    ind_arrays.append(row_inds[colshift::bin_factor,rowshift::bin_factor].reshape(int(np.prod(init_shape)/bin_factor**2),order=self.pixel_order))
                    ind_arrays[-1] = index_map[ind_arrays[-1]][px_mask == True]

            new_data = self.data[ind_arrays[0],:]
            for indarr in ind_arrays[1:]:
                new_data = new_data + self.data[indarr,:]
                
            norm_factor = scipy.sparse.diags(np.ones(self.grid.n_cells)/bin_factor**2,format='csr')
            
            self.data = new_data * norm_factor
            
            self.binning = binning
        

    def set_included_pixels(self,pixel_mask,coords=None):
        '''
        Set which image pixels should be included, or not. Can be 
        used to exclude image pixels which are known to have bad data or 
        otherwise do not conform to the assumptions of the inversion.

        .. note::
            Excluding pixels is a non-reversible process since their
            matrix rows will be removed. It is therefore recommended to keep a
            master copy of the matrix with all pixels included and then
            use this function on a transient copy of the matrix.

        Parameters:

            pixel_mask (numpy.ndarray) : Boolean array the same shape as the un-binned \
                                         camera image, where True or 1 indicates a \
                                         pixel to be included and False or 0 represents \
                                         a pixel to be excluded.

            coords (str)               : Either 'Display' or 'Original', \
                                         specifies what orientation the input \
                                         pixel mask is in. If not givwn, it will be \
                                         auto-detected if possible.
        '''
        if self.pixel_mask is None:
            raise Exception('Cannot set pixel mask for a geometry matrix which does not include full sensor.')
        
        if coords is None:

            # If there are no transform actions, we have no problem.
            if len(self.image_geometry.transform_actions) == 0 and self.image_geometry.pixel_aspectratio == 1:
                coords = self.image_coords

            elif self.image_geometry.get_display_shape() != self.image_geometry.get_original_shape():

                if pixel_mask.shape == self.image_geometry.get_display_shape()[::-1]:
                    coords = 'Display'
                elif pixel_mask.shape == self.image_geometry.get_original_shape()[::-1]:
                    coords = 'Original'
                else:
                    raise ValueError('Input pixel mask has an unexpected shape! Got {:d}x{:d}; expected {:d}x{:d} or {:d}x{:d}'.format(image.shape[1],image.shape[0],self.image_geometry.get_display_shape[1],self.image_geometry.get_display_shape[0],self.image_geometry.get_original_shape[1],self.image_geometry.get_original_shape[0]))
            else:
                raise Exception('Cannot determine mask orientation automatically; please provide the "coords" input argument.')
        
        if coords.lower() == 'display' and self.image_coords.lower() == 'original':

            pixel_mask = self.image_geometry.display_to_original_image(pixel_mask.astype(int),interpolation='nearest')

        elif coords.lower() == 'original' and self.image_coords.lower() == 'display':

            pixel_mask = self.image_geometry.original_to_display_image(pixel_mask.astype(int),interpolation='nearest')


        if self.binning > 1:

            pixel_mask = _bin_image(pixel_mask,self.binning,bin_func=np.min)


        mask_delta = self.pixel_mask.astype(int) - pixel_mask.astype(int)

        if mask_delta.min() == -1:
            raise ValueError('Provided pixel mask includes previously excluded pixels; pixels can only be disabled, not re-enabled using this function.')

        old_px_mask = self.pixel_mask.reshape(self.pixel_mask.size,order=self.pixel_order)
        mask_delta = mask_delta.reshape(mask_delta.size,order=self.pixel_order)

        mask_delta = mask_delta[old_px_mask == True]

        self.data = self.data[mask_delta == 0,:]

        self.pixel_mask = pixel_mask.astype(bool)



    def get_included_pixels(self):
        '''
        Get a mask showing which image pixels are included in the
        geometry matrix.

        Returns:

            numpy.ndarray : Boolean array the same shape as the camera image after binning, \
                            where True indicates the corresponding pixel is included \
                            and False indicates the pixel is excluded.
        '''
        return self.pixel_mask



    def save(self,filename):
        '''
        Save the geometry matrix to a file.

         .. note::
            .npz is the recommended file format; .mat is provided if compatibility
            with MATLAB is required but produces larger file sizes, and .zip is provided
            to make maximally compatible data files but is extremely slow to save and load.

        Parameters:
            
            filename (str) : File name to save to, including file extension. \
                             The file extension determines the format to be saved: \
                             '.npz' for compressed NumPy binary format, \
                             '.mat' for MATLAB format or 
                             '.zip' for Zipped collection of ASCII files.


        '''
        try:
            fmt = filename.split('.')[1:][-1]
        except IndexError:
            raise ValueError('Given file name does not include file extension; extension .npz, .mat or .zip must be included to determine file type!')

        if fmt == 'npz':
            self._save_npz(filename)
        elif fmt == 'mat':
            self._save_matlab(filename)
        elif fmt == 'zip':
            self._save_txt(filename)
        else:
            raise ValueError('File extension "{:s}" not understood; options are "npz", "mat" or "zip".'.format(fmt))



    def format_image(self,image,coords=None):
        '''
        Format a given 2D camera image in to a 1D data vector 
        (i.e. :math:`b` in :math:`Ax = b`) appropriate for use with this 
        geometry matrix. This will bin the image, remove any excluded 
        pixels and reshape it to a 1D vector.

        Parameters: 

            image (numpy.ndarray) : Input image.

            coords (str)          : Either 'Display' or 'Original', \
                                    specifies what orientation the input \
                                    image is in. If not givwn, it will be \
                                    auto-detected if possible.

        Returns:

            scipy.sparse.csr_matrix : 1xN_pixels image data vector. Note that this is \
                                      returned as a sparse matrix object despite its \
                                      density being 100%; this is for consistency with the \
                                      matrix itself.

        '''

        if len(image.shape) > 2:
            raise ValueError('Provided image array has 3 dimensions i.e. includes colour channels. This only supports single channel images.')

        if coords is None:

            # If there are no transform actions, we have no problem.
            if len(self.image_geometry.transform_actions) == 0 and self.image_geometry.pixel_aspectratio == 1:
                coords = self.image_coords

            elif self.image_geometry.get_display_shape() != self.image_geometry.get_original_shape():

                if image.shape == self.image_geometry.get_display_shape()[::-1]:
                    coords = 'Display'
                elif image.shape == self.image_geometry.get_original_shape()[::-1]:
                    coords = 'Original'
                else:
                    raise ValueError('Input image has an unexpected shape! Got {:d}x{:d}; expected {:d}x{:d} or {:d}x{:d}'.format(image.shape[1],image.shape[0],self.image_geometry.get_display_shape[1],self.image_geometry.get_display_shape[0],self.image_geometry.get_original_shape[1],self.image_geometry.get_original_shape[0]))
            else:
                raise Exception('Cannot determine image orientation automatically; please provide the "coords" input argument.')
        
        if coords.lower() == 'display' and self.image_coords.lower() == 'original':

            im_out = self.image_geometry.display_to_original_image(image)

        elif coords.lower() == 'original' and self.image_coords.lower() == 'display':

            im_out = self.image_geometry.original_to_display_image(image)

        else:

            im_out = image.copy()


        if self.binning > 1:
            im_out = _bin_image(im_out,self.binning,bin_func=np.mean)

        elif self.binning < 1:
            raise Exception('This matrix has binning < 1 which is not really meaningful. Set binning =>1 before trying to use this matrix.')

        im_out = im_out.reshape(im_out.size,order=self.pixel_order)

        return scipy.sparse.csr_matrix(im_out[ self.pixel_mask.reshape(self.pixel_mask.size,order=self.pixel_order)])


    def unformat_image(self,im_vector,coords='Native',fill_value=np.nan):
        '''
        Formats a given a 1D data vector of image pixel values (i.e. :math:`b` in :math:`Ax = b`)
        back in to a 2D image array.

        Parameters:

            im_vector (numpy.ndarray or sparse matrix) : 1D vector of image data, must have length equal to the number of rows in the geometry matrix.

            coords (str)                               : What image orientation to return the image: 'Original', 'Display' or 'Native' (whichever was used when creating the geometry matrix).

            fill_value (float)                         : Value to return in any image pixels which are not included in the geometry matrix.

        Returns:

            numpy.ndarray : 2D array containing the image.

        '''
        if self.pixel_mask is None:
            raise Exception('Cannot perform image un-formatting because this GeometryMatrix is not for a complete image area.')

        try:
            im_vector = im_vector.toarray()
        except AttributeError:
            im_vector = np.array(im_vector)

        if (np.array(im_vector.shape) > 1).sum() > 1:
            raise ValueError('Provided image data vector is not 1D! (array shape: {:}). Expected 1D vector.'.format(im_vector.shape))

        im_vector = np.squeeze(im_vector)

        if im_vector.size != self.data.shape[0]:
            raise ValueError('Provided image data vector is not the correct size - expected {:d} values, got {:d}'.format(self.data.shape[0],im_vector.size))

        im_out = np.zeros(self.pixel_mask.shape) + fill_value

        im_out[self.pixel_mask] = im_vector

        if coords.lower() == 'display' and self.image_coords.lower() == 'original':
            im_out = self.image_geometry.original_to_display_image(im_out)
        elif coords.lower() == 'original' and self.image_coords.lower() == 'display':
            im_out = self.image_geometry.display_to_original_image(im_out)

        return im_out


    def _calc_row_volume(self,ray_endpoints):
        '''
        Calculate a matrix row given the sight-line start and end in 3D,
        for volume type grids (i.e. quantities constant within grid cell volume).
        
        Parameters:
            
            ray_endpoints (sequence) : 6-element sequence containing the \
                                       ray start and end coordinates: \
                                       (Xstart, Ystart, Zstart, Xend, Yend, Zend)
            
        Returns:
            
            tuple : Calculated row elements: \
            
                    * numpy.ndarray containing row indices of non-zero elements \
                      in the column.

                    * numpy.ndarray containing the values of the matrix elements at \
                      those row indices.

        '''
        ray_start_coords = np.array(ray_endpoints[:3])
        ray_end_coords = np.array(ray_endpoints[3:])
        
        # Total length of the ray
        ray_length = np.sqrt( np.sum( (ray_end_coords - ray_start_coords)**2 ) )
        
        # Output will be stored in here
        #row_out = scipy.sparse.lil_matrix( (1,self.grid.n_cells) )
        col_inds = np.arange(self.grid.n_cells,dtype=np.uint32)
        data = np.zeros(self.grid.n_cells)
        
        # Get the ray intersections with the grid cells
        positions,intersected_cells = self.grid.get_cell_intersections(ray_start_coords,ray_end_coords)

        # Convert the lists of intersected cells in to sets for later convenience.
        intersected_cells = [set(cells) for cells in intersected_cells]

        # For keeping track of which cell we're currently in
        in_cell = set()
        
        # Loop over each intersection
        for i in range(positions.size):
            
            if len(in_cell) == 0:
                # Entering the grid
                in_cell = intersected_cells[i]
                
            else:
                # Going from one cell to another         
                leaving_cell = list(intersected_cells[i] & in_cell)
                
                if len(leaving_cell) == 1:
                    
                    #row_out[0,leaving_cell[0]] = row_out[0,leaving_cell[0]] + (positions[i] - positions[i-1])
                    data[leaving_cell[0]] = data[leaving_cell[0]] + (positions[i] - positions[i-1])
                    in_cell = intersected_cells[i]
                    in_cell.remove(leaving_cell[0])
                
                else:
                    '''
                    # Make a plot to show where on the grid the error has occured.
                    self.grid.plot()
                    self.grid.get_cell_intersections(ray_start_coords,ray_end_coords,plot=True)
                    problem_point = ray_start_coords + positions[i]*(ray_end_coords - ray_start_coords)/ray_length
                    plt.plot(np.sqrt(problem_point[0]**2 + problem_point[1]**2),problem_point[2],'bo')
                    plt.title('Error for: {:},{:}'.format(ray_start_coords,ray_end_coords))
                    plt.show()
                    '''
                    raise Exception('Error generating geometry matrix row: could not identify which grid cell the LoS left.')
        
        
        # If the sight line ends inside a cell, add the length it was inside that cell.
        if len(in_cell) > 0:
            
            leaving_cell = list(in_cell)
            
            if len(leaving_cell) == 1:
                data[leaving_cell[0]] = data[leaving_cell[0]] + (ray_length - positions[-1])
            else:
                raise Exception('Error generating geometry matrix row: could not identify which grid cell the LoS left.')

        return col_inds[data > 0],data[data > 0]



    def _save_npz(self,filename):
        '''
        Save the geometry matrix in compressed NumPy binary format.
        '''
        coo_data = self.data.tocoo()
        
        np.savez_compressed( filename,
                             mat_row_inds = coo_data.row,
                             mat_col_inds = coo_data.col,
                             mat_data = coo_data.data,
                             mat_shape = self.data.shape,
                             grid_verts = self.grid.vertices,
                             grid_cells = self.grid.cells,
                             grid_wall = self.grid.wall_contour,
                             binning = self.binning,
                             pixel_order = self.pixel_order,
                             pixel_mask = self.pixel_mask,
                             history = self.history,
                             grid_type = self.grid.__class__.__name__,
                             im_transforms = self.image_geometry.transform_actions,
                             im_px_aspect = self.image_geometry.pixel_aspectratio,
                             im_shape = self.pixel_mask.shape[::-1],
                             im_coords = self.image_coords
                            )

    def _load_npz(self,filename):
        '''
        Load a geometry matrix from a compressed NumPy binary file
        '''
        f = np.load(filename, allow_pickle=True)
        
        self.binning = float(f['binning'])
        self.pixel_order = str(f['pixel_order'])
        self.history = f['history'].item()
        self.pixel_mask = f['pixel_mask']
        self.image_coords = str(f['im_coords'])
        self.image_geometry = CoordTransformer()
        self.image_geometry.set_transform_actions(f['im_transforms'])
        self.image_geometry.set_pixel_aspect(f['im_px_aspect'],relative_to='Original')
        self.image_geometry.set_image_shape(*self.binning*np.array(self.pixel_mask.shape[::-1]),coords=self.image_coords)
        
        self.grid = PoloidalVolumeGrid(f['grid_verts'],f['grid_cells'],f['grid_wall'],src=self.history['grid'])
        self.data = scipy.sparse.csr_matrix((f['mat_data'],(f['mat_row_inds'],f['mat_col_inds'])),shape=f['mat_shape'])



    def _save_matlab(self,filename):
        '''
        Save the geometry matrix in MATLAB format.
        '''
        scipy.io.savemat( filename,
                         { 'geom_mat': self.data,
                           'grid_verts':self.grid.vertices,
                           'grid_cells':self.grid.cells,
                           'grid_wall':self.grid.wall_contour,
                           'binning':self.binning,
                           'pixel_order':self.pixel_order,
                           'sightline_history':self.history['los'],
                           'matrix_history':self.history['matrix'],
                           'grid_history':self.history['grid'],
                           'pixel_mask':self.pixel_mask,
                           'grid_type':self.grid.__class__.__name__,
                           'im_transforms': np.array(self.image_geometry.transform_actions,dtype=np.object),
                           'im_px_aspect':self.image_geometry.pixel_aspectratio,
                           'im_coords':self.image_coords
                         }
                        )
        
        
    def _load_matlab(self,filename):
        '''
        Load geometry matrix from a MATLAB file.
        '''
        f = scipy.io.loadmat(filename)
        
        
        
        self.binning = float(f['binning'][0,0])
        self.pixel_order = str(f['pixel_order'][0])
        self.pixel_mask = f['pixel_mask']
        self.history = {'los':str(f['sightline_history'][0]),'grid':str(f['grid_history'][0]),'matrix':str(f['matrix_history'][0])}
        self.image_coords = str(f['im_coords'][0])

        self.grid = PoloidalVolumeGrid(f['grid_verts'],f['grid_cells'],f['grid_wall'],src=self.history['grid'])
        self.data = f['geom_mat'].tocsr()
        
        self.image_geometry = CoordTransformer()
        self.image_geometry.set_transform_actions(f['im_transforms'])
        self.image_geometry.set_pixel_aspect(f['im_px_aspect'],relative_to='Original')
        self.image_geometry.set_image_shape(*self.binning*np.array(self.pixel_mask.shape[::-1]),coords=self.image_coords)


    def _save_txt(self,filename):
        '''
        Save the geometry matrix as a set of zipped ASCII files.
        '''
        coo_data = self.data.tocoo()
        
        with ZipSaveFile(filename,'w') as zfile:
            
            dest = zfile.get_temp_path()
            np.savetxt(os.path.join(dest,'mat_row_ind.txt'),coo_data.row,fmt='%d')
            np.savetxt(os.path.join(dest,'mat_col_ind.txt'),coo_data.col,fmt='%d')
            np.savetxt(os.path.join(dest,'mat_data.txt'),coo_data.data,fmt='%.5e')
            np.savetxt(os.path.join(dest,'grid_verts.txt'),self.grid.vertices,fmt='%.5e')
            np.savetxt(os.path.join(dest,'grid_cells.txt'),self.grid.cells,fmt='%d')
            np.savetxt(os.path.join(dest,'grid_wall.txt'),self.grid.wall_contour)
            np.savetxt(os.path.join(dest,'pixel_mask.txt'),self.pixel_mask,fmt='%d')
            
            meta = {'mat_shape':self.data.shape, 'binning':self.binning, 'pixel_order':self.pixel_order, 'grid_type':self.grid.__class__.__name__,'history':self.history,'im_transform_actions':self.image_geometry.transform_actions,'im_px_aspect':self.image_geometry.pixel_aspectratio,'im_coords':self.image_coords}
            
            with zfile.open_file('metadata.json','w') as metafile:
                
                json.dump(meta,metafile,indent=4)
            
    
    def _load_txt(self,filename):
        '''
        Load a geometry matrix from a set of zipped ASCII files.
        '''
        with ZipSaveFile(filename,'r') as zfile:
            
            src = zfile.get_temp_path()

            
            with zfile.open_file('metadata.json','r') as metafile:
                
                meta = json.load(metafile)

            verts = np.loadtxt(os.path.join(src,'grid_verts.txt'))
            cells = np.loadtxt(os.path.join(src,'grid_cells.txt'),dtype=np.uint32)
            wall = np.loadtxt(os.path.join(src,'grid_wall.txt'))

            self.image_coords = meta['im_coords']

            self.pixel_mask = np.loadtxt(os.path.join(src,'pixel_mask.txt')).astype(bool)
            
            
                
            self.binning = meta['binning']
            self.pixel_order = meta['pixel_order']
            self.history = meta['history']
            
            row_ind = np.loadtxt(os.path.join(src,'mat_row_ind.txt'),dtype=np.uint32)
            col_ind = np.loadtxt(os.path.join(src,'mat_col_ind.txt'),dtype=np.uint32)
            data = np.loadtxt(os.path.join(src,'mat_data.txt'))

            self.grid = PoloidalVolumeGrid(verts,cells,wall,meta['history']['grid'],src=self.history['grid'])
            self.data = scipy.sparse.csr_matrix((data,(row_ind,col_ind)),shape=meta['mat_shape'])

            self.image_geometry = CoordTransformer()
            self.image_geometry.set_transform_actions(meta['im_transforms'])
            self.image_geometry.set_pixel_aspect(meta['im_px_aspect'],relative_to='Original')
            self.image_geometry.set_image_shape(*self.binning*np.array(self.pixel_mask.shape[::-1]),coords=self.image_coords)


    def __str__(self):
        '''
        Get a pretty string describing the matrix in great detail.
        '''
        msg = '======================\nCalcam Geometry Matrix\n======================\n\n'

        msg = msg + self.history['matrix'] + '\n'
        msg = msg + 'Matrix Dimensions: {:d} x {:d} [lines-of-sight x grid elements]\n'.format(*self.data.shape)
        msg = msg + 'Density:           {:.1f}%\n\n'.format(100*float(self.data.getnnz(None)/np.prod(self.data.shape)))

        msg = msg + '--------------\nLines of Sight\n--------------\n'
        msg = msg + self.history['los'] + '\n'

        if self.image_coords.lower() == 'original':
        	im_shape = self.image_geometry.get_original_shape()
        elif self.image_coords.lower() == 'display':
        	im_shape = self.image_geometry.get_display_shape()

        msg = msg + 'Image dimensions:  {:d} x {:d} px\n'.format(im_shape[0],im_shape[1])
        msg = msg + 'Image orientation: {:s}\n'.format(self.image_coords)
        msg = msg + 'Image binning:     {:.1f} x {:.1f}\n\n'.format(self.binning,self.binning)
        

        msg = msg + '-------------------\nReconstruction Grid\n-------------------\n'
        msg = msg + self.history['grid'] + '\n'
        extent = self.grid.extent
        msg = msg + 'Type:                   {:s}\n'.format(self.grid.gridtype)
        msg = msg + 'R coverage:             {:.3f}m - {:.3f}m\n'.format(extent[0],extent[1])
        msg = msg + 'Z coverage:             {:.3f}m - {:.3f}m\n'.format(extent[2],extent[3])

        return msg


    @classmethod
    def fromfile(cls,filename):
        '''
        Load a saved geometry matrix from disk.
        
        Parameters:
            
            filename (str)  : File name to load from. Can be a NumPy (.npz), MATLAB (.mat) or \
                              zipped ASCII (.zip) file.
                             
        Returns:
            
            calcam.GeometryMatrix : Loaded geometry matrix.
            
        '''
        geommat = cls(None,None)
        
        try:
            fmt = filename.split('.')[1:][-1]
        except IndexError:
            raise ValueError('Given file name does not include file extension; extension must be specified to determine file type!')

        if fmt == 'npz':
            geommat._load_npz(filename)
        elif fmt == 'mat':
            geommat._load_matlab(filename)
        elif fmt == 'zip':
            geommat._load_txt(filename)
        else:
            raise ValueError('File extension "{:s}" not understood; should be an "npz", "mat" or "zip" file.'.format(fmt))

        return geommat
        

def squaregrid(wall_contour,cell_size,rmin=None,rmax=None,zmin=None,zmax=None):
    '''
    Create a reconstruction grid with square grid cells.
    
    Parameters:
        
        wall_contour (str or numpy.ndarray) : Either the name of a Calcam CAD model \
                                              from which to use the wall contour, or an \
                                              N x 2 array of R,Z points defining the machine wall.
                                           
        cell_size (float)                   : Side length of each grid cell in metres.
        
        floats rmin, rmax, zmin, zmax       : Optional limits of the grid extent in the R, Z plane. \
                                              Any combination of these may or may not be given; if none \
                                              are given the entire wall contour interior is gridded.
                                           
    Returns:
        
        calcam.PoloidalVolumeGrid    : Generated grid.
        
    '''
    
    # If given a machine name for the wall contour, get the R,Z contour
    if type(wall_contour) is str:
        wall_contour = _get_ccm_wall(wall_contour)

    # If we have a wall contour which joins back up with itself, remove
    # the last point.
    if np.abs(wall_contour[0,:] - wall_contour[-1,:]).max() < 1e-15:
        wall_contour = wall_contour[:-1]


    if rmin is None:
        rmin = wall_contour[:,0].min()
    else:
        rmin = max(rmin,wall_contour[:,0].min())

    if rmax is None:
        rmax = wall_contour[:,0].max()
    else:
        rmax = min(rmax,wall_contour[:,0].max())

    if zmin is None:
        zmin = wall_contour[:,1].min()
    else:
        zmin = max(zmin,wall_contour[:,1].min())

    if zmax is None:
        zmax = wall_contour[:,1].max() 
    else:
        zmax = min(zmax,wall_contour[:,1].max())


    Rmax = cell_size * np.ceil( (rmax - rmin) / cell_size ) + rmin
    Zmax = cell_size * np.ceil( (zmax - zmin) / cell_size ) + zmin

    nr = int( (Rmax - rmin)/cell_size) + 1
    nz = int( (Zmax - zmin)/cell_size) + 1

    Rpts = np.linspace(rmin,Rmax,nr)
    Zpts = np.linspace(zmin,Zmax,nz)

    wall_path = mplpath.Path(np.vstack((wall_contour,wall_contour[-1,:])),closed=True)

    vertices = np.zeros((nr * nz,2))
    cells = np.zeros( ( (nr-1)*(nz-1) , 4 ),dtype=np.uint32 )

    cell_ind = 0
    vert_ind = 0

    _verts = np.zeros((4,2))

    for iz in range(Zpts.size-1):
        for ir in range(Rpts.size-1):

            # 4 corners of the square grid cell
            _verts[0,:] = [Rpts[ir],Zpts[iz]]
            _verts[1,:] = [Rpts[ir+1],Zpts[iz]]
            _verts[2,:] = [Rpts[ir+1],Zpts[iz+1]]
            _verts[3,:] = [Rpts[ir],Zpts[iz+1]]

            # If the cell is completely outside the wall, don't bother.
            if not np.any(wall_path.contains_points(_verts)):
                continue

            for vert in range(4):
                try:
                    cells[cell_ind][vert] = np.where( (vertices[:,0] == _verts[vert,0]) & (vertices[:,1] == _verts[vert,1]) )[0][0]
                except IndexError:
                    vertices[vert_ind,:] = _verts[vert,:]
                    cells[cell_ind][vert] = vert_ind
                    vert_ind += 1
            cell_ind += 1

    
    return PoloidalVolumeGrid(vertices[:vert_ind,:],cells[:cell_ind,:],wall_contour,src='Square grid with {:.1f}cm cells generated using squaregrid()'.format(cell_size*1e2))




def trigrid(wall_contour,max_cell_scale,rmin=None,rmax=None,zmin=None,zmax=None,**kwargs):
    '''
    Create a reconstruction grid with triangular grid cells conforming to the 
    wall contour. Requires the MeshPy package (https://pypi.org/project/MeshPy/) 
    to be installed, since it uses J. Shewchuk's "triangle" via the MeshPy to 
    generate the grid.
    
    Parameters:
        
        wall_contour (str or numpy.ndarray) : Either the name of a Calcam CAD model \
                                              from which to use the wall contour, or an \
                                              N x 2 array of R,Z points defining the machine wall.
                                           
        max_cell_scale (float)              : Approximate maximum allowed length scale for a grid \
                                              cell, in metres.
        
        floats rmin, rmax, zmin, zmax       : Optional limits of the grid extent in the R, Z plane. \
                                              Any combination of these may or may not be given; if none \
                                              are given the entire wall contour interior is gridded.
                                           
    Any additional keyword arguments will be passed directly to the triangle mesher, meshpy.triangle.build().
    This can be used to further control the meshing; see the MeshPy documentation for available arguments.
    
                                           
    Returns:
        
        calcam.PoloidalVolumeGrid    : Generated grid.
        
    '''
    # Before trying to generate a triangular grid, check we have the meshpy package available.
    if meshpy_err is not None:
        raise Exception('This function requires the MeshPy module (https://pypi.org/project/MeshPy), which could not be imported: {:s}'.format(meshpy_err))

    # If given a machine name for the wall contour, get the R,Z contour
    if type(wall_contour) is str:
        wall_contour = _get_ccm_wall(wall_contour)

    # If we have a wall contour which joins back up with itself, remove
    # the last point.
    if np.abs(wall_contour[0,:] - wall_contour[-1,:]).max() < 1e-15:
        wall_contour = wall_contour[:-1]


    # Whatever happens, the wall will define some vertices to pass to the mesher.
    verts = wall_contour

    # Connectivity between wall contour vertices is simply each one
    # connected to the next, then the last connected to the first
    segments = np.zeros((verts.shape[0],2),dtype=np.uint32)
    segments[:,0] = np.arange(verts.shape[0])
    segments[:,1] = np.arange(1,verts.shape[0]+1)
    segments[-1][-1] = 0

    max_cell_area = (max_cell_scale**2) * np.sqrt(2)

    if not all( [limit is None for limit in [rmin,rmax,zmin,zmax] ] ):

        epsilon = 2*np.sqrt(max_cell_area)
        if rmin is None:
            rmin = wall_contour[:,0].min() - epsilon
        else:
            rmin = max(rmin,wall_contour[:,0].min() - epsilon)

        if rmax is None:
            rmax = wall_contour[:,0].max() + epsilon
        else:
            rmax = min(rmax,wall_contour[:,0].max() + epsilon)

        if zmin is None:
            zmin = wall_contour[:,1].min() - epsilon
        else:
            zmin = max(zmin,wall_contour[:,1].min() - epsilon)

        if zmax is None:
            zmax = wall_contour[:,1].max() + epsilon
        else:
            zmax = min(zmax,wall_contour[:,1].max() + epsilon)

        bbox_verts = np.zeros((4,2))
        bbox_verts[0,:] = [rmin,zmin]
        bbox_verts[1,:] = [rmax,zmin]
        bbox_verts[2,:] = [rmax,zmax]
        bbox_verts[3,:] = [rmin,zmax]

        bbox_segments = np.array( [ [0,1] , [1,2] , [2,3] , [3,0] ] ) + verts.shape[0] 

        epsilon = epsilon / 2.

        holeR = np.linspace(rmin - epsilon,rmax + epsilon, int((2*epsilon+rmax-rmin)/epsilon))
        holeZ = np.linspace(zmin - epsilon,zmax + epsilon, int((zmax-zmin)/epsilon))
        
        holeR,holeZ = np.meshgrid(holeR,holeZ)
        holecoords = np.hstack( (holeR.reshape(holeR.size,1), holeZ.reshape(holeZ.size,1)) )

        wall_path = mplpath.Path(np.vstack((verts,verts[-1,:])),closed=True)
        bbox_path = mplpath.Path(np.vstack((bbox_verts,bbox_verts[-1,:])),closed=True)

        inds =  np.logical_xor( bbox_path.contains_points(holecoords) , wall_path.contains_points(holecoords) )
        holecoords = holecoords[inds,:]

        geometry = {'vertices' : np.vstack( (verts,bbox_verts) ) , 'segments' : np.vstack( (segments,bbox_segments) ), 'holes': holecoords}

    else:

        geometry = {'vertices': verts, 'segments' : segments}


    mesher_kwargs = dict(kwargs)
    mesher_kwargs['max_volume'] = max_cell_area

    # Run triangle to make the grid!
    # Do this in a separate process so that if the triangle library has a fail,
    # it doesn't drag this whole python process down with it.
    q = multiprocessing.Queue()
    triprocess = multiprocessing.Process(target=_run_triangle_python,args=(geometry,q),kwargs=mesher_kwargs)
    triprocess.start()
    while triprocess.is_alive() and q.empty():
        time.sleep(0.05)
    
    if not q.empty():
        mesh = q.get()
        if isinstance(mesh,Exception):
            raise mesh
    else:
        raise Exception('Call to meshpy.triangle to generate the grid failed silently. The triangle library can be a bit flaky; maybe try again and/or change settings.')

    return PoloidalVolumeGrid(mesh['vertices'],mesh['cells'],wall_contour,src='Wall-conforming triangular mesh with ~{:.1f}cm cells generated with trigrid()'.format(max_cell_scale*1e2))




def _get_ccm_wall(model_name):
    '''
    Get the wall contour points for a named machine, if Calcam has a 
    CAD model for the machine with a wall contour in it.
    
    Parameters:
        
        model_name (str) : Name of the Calcam CAD model to get the wall for.
        
    Returns:
        
        numpy.ndarray : Nx2 array of (R,Z) coordinates around the wall contour
        
    '''
    cadmodels = config.CalcamConfig().get_cadmodels()
    if model_name not in cadmodels.keys():
        raise ValueError('Unknown name "{:s}" for wall contour; available machine names are: {:s}'.format(model_name,', '.join(cadmodels.keys())))
    else:
        with ZipSaveFile(cadmodels[model_name][0],'rs') as deffile:
            # Load the wall contour, if present
            if 'wall_contour.txt' in deffile.list_contents():
                with deffile.open_file('wall_contour.txt','r') as cf:
                    wall_contour = np.loadtxt(cf)
            else:
                raise Exception('CAD model "{:s}" does not define a wall contour!'.format(model_name))

    return wall_contour



def _run_triangle_python(geometry,queue,**kwargs):
    '''
    Run triangular mesh generation with the triangle library via MeshPy module.
    See MeshPy documentation.

    This function is designed to be used from a multiprocessing call, to avoid
    errors in the underlying triangle code from dragging down the main Python process
    (seen on Windows 10 / Py3.5.)

    Parameters:

        geometry (dict)               : Dictionary defining the bounding geometry, see triangle \
                                        manual.

        queue (multiprocessing.Queue) : Multiprocessing queue on which to place the result.
        
    Any keyword arguments are all passed to meshpy.triangle.build()

    '''
    # Re-arrange the input geometry and mesher arguments. This
    # is all done here because MeshPy crashes silently otherwise,
    # it appears to be highly not process-safe.

    # Put the input geometry in to a meshpy object
    boundary_geom = meshpy.triangle.MeshInfo()
    boundary_geom.set_points(geometry['vertices'])
    boundary_geom.set_facets(geometry['segments'])
    if 'holes' in geometry:
        boundary_geom.set_holes(geometry['holes'])

    # Don't just pass the kwargs straight to MeshPy or it crashes.
    mesher_args = dict(kwargs)

    try:
        # Run triangle and send the results back to the calling thread.
        mesh = meshpy.triangle.build(boundary_geom,**mesher_args)
        queue.put( {'vertices':np.array(mesh.points), 'cells':np.array(mesh.elements)}  )
    except Exception as e:
        queue.put(e)




def _bin_image(image,bin_factor,bin_func=np.mean):

    newshape = ( image.shape[0] // bin_factor, bin_factor, image.shape[1] // bin_factor, bin_factor )
    newshape = np.array(newshape,dtype=int)
    return bin_func( bin_func( image.reshape(newshape),axis=3 ) ,axis=1 )



'''
* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

import sys
try:
    from calcam import gui
except:
    gui = False

if __name__ == '__main__':

    if not gui:
        print('Could not start Calcam GUI.')
        exit()
        
    try:
        arg = sys.argv[1]
    except:
        arg = '--launcher'

    if arg == '--launcher':
        gui.open_window(gui.Launcher)
    elif arg == '--fitting_calib':
        gui.open_window(gui.FittingCalib)
    elif arg == '--alignment_calib':
        gui.open_window(gui.AlignmentCalib)
    elif arg == '--virtual_calib':
        gui.open_window(gui.VirtualCalib)
    elif arg == '--viewer':
        gui.open_window(gui.Viewer)
    elif arg == '--image_analyser':
        gui.open_window(gui.ImageAnalyser)
    elif arg == '--settings':
        gui.open_window(gui.Settings)
    elif arg == '--cad_edit':
        try:
            filepath = sys.argv[2]
        except:
            filepath = None
        gui.open_window(gui.CADEdit,filepath)
